from functools import partial
from typing import Optional

from amulet_editor.application import appearance
from amulet_editor.application.appearance import Theme
from amulet_editor.data import packages, project
from amulet_editor.models.package import AmuletTool
from amulet_editor.models.widgets import ADragContainer, ATooltipIconButton
from amulet_editor.tools.packages import Packages
from amulet_editor.tools.settings import Settings
from amulet_editor.tools.startup import Startup
from PySide6.QtCore import QCoreApplication, QRect, QSize, Qt
from PySide6.QtGui import QAction, QKeyEvent, QMouseEvent, QResizeEvent
from PySide6.QtWidgets import (
    QApplication,
    QButtonGroup,
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QMenu,
    QMenuBar,
    QSizePolicy,
    QSplitter,
    QStackedWidget,
    QStatusBar,
    QVBoxLayout,
    QWidget,
)

MINIMUM_PANEL_WIDTH = 250
DEFAULT_PANEL_WIDTH = 300


class AmuletWindow(QMainWindow):
    def __init__(self) -> None:
        super().__init__()

        self.setupUi()

        # Setup side menu
        self.btng_menus = QButtonGroup(self)
        self.active_tool = Startup()
        self.secondary_panel_none = QWidget()

        self.load_tool(self.active_tool)
        self.load_tool(Packages(), True)
        self.load_tool(Settings(), True)

        self.btng_menus.buttons()[0].setChecked(True)
        self.show_tool(self.active_tool)

        # Connect signals
        self.spl_horizontal.splitterMoved.connect(self.update_panel_sizes)

        # Connect restyle signal and apply current theme
        appearance.changed.connect(self.theme_changed)
        project.changed.connect(self.reload_project)

        # Connect menu options
        self.act_show_primary_panel.setCheckable(True)
        self.act_show_primary_panel.setChecked(True)
        self.act_show_primary_panel.triggered.connect(self.toggle_primary_panel)
        self.act_show_secondary_panel.setCheckable(True)
        self.act_show_secondary_panel.setChecked(False)
        self.act_show_secondary_panel.triggered.connect(self.toggle_secondary_panel)
        self.act_close_project.triggered.connect(self.close_project)

        for theme_name in appearance.list_themes():
            action = QAction(self.mn_appearance)

            action.setCheckable(True)
            if theme_name == appearance.theme().name:
                action.setChecked(True)
            action.setText(theme_name)

            action.triggered.connect(partial(appearance.set_theme, theme_name))

            self.mn_appearance.addAction(action)

    def close_project(self) -> None:
        self.unload_tools()

        startup_tool = Startup()
        self.load_tool(startup_tool)
        self.show_tool(startup_tool)

        self.act_close_project.setEnabled(False)

    def reload_project(self, project_root: str) -> None:
        """
        Unload all tools currently in use and replace them with new instances.
        """
        self.unload_tools()

        tools = packages.list_tools()
        for tool in tools:
            self.load_tool(tool)

        self.show_tool(tools[0])

        self.act_close_project.setEnabled(True)

    def update_panel_sizes(self, distance: int, index: int) -> None:
        """
        Detect whether both panels can fit on screen and determine which panel should be collapsed if there is only space for one.
        """
        page_width = self.spl_horizontal.width()

        if not self.swgt_primary_panel.visibleRegion().isEmpty():
            page_width = page_width - self.swgt_primary_panel.width()

        if not self.swgt_secondary_panel.visibleRegion().isEmpty():
            page_width = page_width - self.swgt_secondary_panel.width()

        if page_width < self.swgt_pages.minimumWidth():
            if index == 1:
                self.act_show_secondary_panel.setChecked(False)
                self.toggle_secondary_panel()
            elif index == 2:
                self.act_show_primary_panel.setChecked(False)
                self.toggle_primary_panel()

    def toggle_primary_panel(self) -> None:
        """
        Detect primary panel view state and toggle panel to match menu option.
        """
        if (
            self.act_show_primary_panel.isChecked()
            and self.swgt_primary_panel.visibleRegion().isEmpty()
        ):
            if not self.swgt_secondary_panel.visibleRegion().isEmpty():
                secondary_panel_width = self.swgt_secondary_panel.width()
                page_width = (
                    self.spl_horizontal.width()
                    - secondary_panel_width
                    - DEFAULT_PANEL_WIDTH
                )
                if page_width < self.swgt_pages.minimumWidth():
                    secondary_panel_width = 0
                    page_width = self.spl_horizontal.width() - DEFAULT_PANEL_WIDTH
                    self.act_show_secondary_panel.setChecked(False)
            else:
                secondary_panel_width = 0
                page_width = self.spl_horizontal.width() - DEFAULT_PANEL_WIDTH

            self.spl_horizontal.setSizes(
                [DEFAULT_PANEL_WIDTH, page_width, secondary_panel_width]
            )
        elif not self.swgt_primary_panel.visibleRegion().isEmpty():
            secondary_panel_width = (
                self.swgt_secondary_panel.width()
                if self.act_show_secondary_panel.isChecked()
                else 0
            )
            page_width = self.spl_horizontal.width() - secondary_panel_width
            self.spl_horizontal.setSizes([0, page_width, secondary_panel_width])

    def toggle_secondary_panel(self) -> None:
        """
        Detect secondary panel view state and toggle panel to match menu option.
        """
        if (
            self.act_show_secondary_panel.isChecked()
            and self.swgt_secondary_panel.visibleRegion().isEmpty()
        ):
            if not self.swgt_primary_panel.visibleRegion().isEmpty():
                primary_panel_width = self.swgt_primary_panel.width()
                page_width = (
                    self.spl_horizontal.width()
                    - primary_panel_width
                    - DEFAULT_PANEL_WIDTH
                )
                if page_width < self.swgt_pages.minimumWidth():
                    primary_panel_width = 0
                    page_width = self.spl_horizontal.width() - DEFAULT_PANEL_WIDTH
                    self.act_show_primary_panel.setChecked(False)
            else:
                primary_panel_width = 0
                page_width = self.spl_horizontal.width() - DEFAULT_PANEL_WIDTH

            self.spl_horizontal.setSizes(
                [primary_panel_width, page_width, DEFAULT_PANEL_WIDTH]
            )
        elif not self.swgt_secondary_panel.visibleRegion().isEmpty():
            primary_panel_width = (
                self.swgt_primary_panel.width()
                if self.act_show_primary_panel.isChecked()
                else 0
            )
            page_width = self.spl_horizontal.width() - primary_panel_width
            self.spl_horizontal.setSizes([primary_panel_width, page_width, 0])

    def load_tool(self, tool: AmuletTool, static: bool = False) -> None:
        """
        Loads tool into the main application toolbar.
        Tools marked as static are considered core to the functionality of the editor and thus cannot be removed once loaded.
        """
        icon_button = ATooltipIconButton(tool.icon_name, self)
        icon_button.clicked.connect(partial(self.show_tool, tool))
        icon_button.setCheckable(True)
        icon_button.setFixedSize(QSize(40, 40))
        icon_button.setToolTip(tool.name)
        icon_button.setIconSize(QSize(30, 30))

        self.btng_menus.addButton(icon_button)

        if static:
            self.lyt_static_tools.addWidget(icon_button)
        else:
            self.wgt_dynamic_tools.add_item(icon_button)

        tool.page.changed.connect(partial(self.change_page, tool))
        tool.primary_panel.changed.connect(partial(self.change_primary_panel, tool))
        tool.secondary_panel.changed.connect(partial(self.change_secondary_panel, tool))

    def unload_tools(self) -> None:
        # TODO: This should also clear corresponding panels and pages
        for index in reversed(range(self.wgt_dynamic_tools.layout().count())):
            self.wgt_dynamic_tools.layout().itemAt(index).widget().deleteLater()

    def show_tool(self, tool: AmuletTool) -> None:
        self.active_tool = tool
        self.lbl_primary_panel_title.setText(tool.name)
        self.show_page(tool.page.widget())
        self.show_primary_panel(tool.primary_panel.widget())
        self.show_secondary_panel(tool.secondary_panel.widget())

    def show_page(self, page: QWidget) -> None:
        if page not in self.swgt_pages.children():
            self.swgt_pages.addWidget(page)
        self.swgt_pages.setCurrentWidget(page)

    def show_primary_panel(self, panel: QWidget) -> None:
        if panel not in self.swgt_primary_panel.children():
            self.swgt_primary_panel.addWidget(panel)
        self.swgt_primary_panel.setCurrentWidget(panel)

    def show_secondary_panel(self, panel: Optional[QWidget]) -> None:
        if panel is None:
            panel = self.secondary_panel_none
        if panel not in self.swgt_secondary_panel.children():
            self.swgt_secondary_panel.addWidget(panel)
        self.swgt_secondary_panel.setCurrentWidget(panel)

        if (
            panel is not self.secondary_panel_none
            and self.swgt_secondary_panel.visibleRegion().isEmpty()
        ):
            self.act_show_secondary_panel.setChecked(True)
            self.toggle_secondary_panel()

    def change_page(self, tool: AmuletTool, widget: QWidget) -> None:
        if self.active_tool == tool:
            self.show_page(widget)

    def change_primary_panel(self, tool: AmuletTool, widget: QWidget) -> None:
        if self.active_tool == tool:
            self.show_primary_panel(widget)

    def change_secondary_panel(
        self, tool: AmuletTool, widget: Optional[QWidget]
    ) -> None:
        if self.active_tool == tool:
            self.show_secondary_panel(widget)

    def theme_changed(self, theme: Theme) -> None:
        for theme_action in self.mn_appearance.actions():
            if theme_action.text() == theme.name:
                theme_action.setChecked(True)
            else:
                theme_action.setChecked(False)

    def resizeEvent(self, event: QResizeEvent) -> None:
        self.update_panel_sizes(MINIMUM_PANEL_WIDTH, 1)

        return super().resizeEvent(event)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        focused = QApplication.focusWidget()
        if event.key() == Qt.Key_Escape and focused is not None:
            focused.clearFocus()

        return super().keyPressEvent(event)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        focused = QApplication.focusWidget()
        if not focused is None:
            focused.clearFocus()

        return super().mousePressEvent(event)

    def setupUi(self):
        # Create 'Application' widget
        self.wgt_application = QWidget(self)

        # Create 'Application' layout
        self.lyt_application = QGridLayout(self.wgt_application)

        # Configure frame for 'Static Menu' items
        self.frm_static_tools = QFrame(self.wgt_application)
        self.frm_static_tools.setFixedWidth(40)

        # Configure widget for 'Dynamic Menu' items
        self.wgt_dynamic_tools = ADragContainer(self.frm_static_tools)
        self.wgt_dynamic_tools.layout().setAlignment(Qt.AlignTop)
        self.wgt_dynamic_tools.layout().setContentsMargins(0, 0, 0, 0)
        self.wgt_dynamic_tools.layout().setSpacing(0)

        # Configure 'Static Menu' layout
        self.lyt_static_tools = QVBoxLayout(self.frm_static_tools)
        self.lyt_static_tools.setSpacing(0)
        self.lyt_static_tools.setContentsMargins(0, 0, 0, 0)
        self.lyt_static_tools.addWidget(self.wgt_dynamic_tools)

        # Create splitter for 'Application' page and panels
        self.spl_horizontal = QSplitter(self.wgt_application)

        # Configure 'Primary Panel' frame
        self.frm_primary_panel = QFrame(self.spl_horizontal)
        self.frm_primary_panel.setFrameShape(QFrame.NoFrame)
        self.frm_primary_panel.setFrameShadow(QFrame.Raised)

        # Create 'Primary Panel' layout
        self.lyt_primary_panel = QVBoxLayout(self.frm_primary_panel)

        # Configure 'Primary Panel' header
        self.frm_primary_panel_header = QFrame(self.frm_primary_panel)
        self.frm_primary_panel_header.setMinimumSize(QSize(0, 25))
        self.frm_primary_panel_header.setFrameShape(QFrame.NoFrame)
        self.frm_primary_panel_header.setFrameShadow(QFrame.Raised)

        # Create title for 'Primary Panel'
        self.lbl_primary_panel_title = QLabel(self.frm_primary_panel_header)

        # Configure 'Primary Panel' header layout
        self.lyt_primary_panel_header = QHBoxLayout(self.frm_primary_panel_header)
        self.lyt_primary_panel_header.addWidget(self.lbl_primary_panel_title)
        self.lyt_primary_panel_header.setSpacing(0)
        self.lyt_primary_panel_header.setContentsMargins(9, 0, 9, 0)

        # Configure 'Primary Panel' stacked widget (container for 'Primary Panel' widgets)
        self.swgt_primary_panel = QStackedWidget(self.frm_primary_panel)
        spol_pkg_primary_panel = QSizePolicy(QSizePolicy.Ignored, QSizePolicy.Preferred)
        spol_pkg_primary_panel.setHorizontalStretch(0)
        spol_pkg_primary_panel.setVerticalStretch(0)
        self.swgt_primary_panel.setSizePolicy(spol_pkg_primary_panel)
        self.swgt_primary_panel.setMinimumSize(QSize(MINIMUM_PANEL_WIDTH, 0))

        # Configure 'Primary Panel' layout
        self.lyt_primary_panel.addWidget(self.frm_primary_panel_header)
        self.lyt_primary_panel.addWidget(self.swgt_primary_panel)
        self.lyt_primary_panel.setSpacing(0)
        self.lyt_primary_panel.setContentsMargins(0, 0, 0, 0)

        # Configure 'Secondary Panel' frame
        self.frm_secondary_panel = QFrame(self.spl_horizontal)
        self.frm_secondary_panel.setFrameShape(QFrame.NoFrame)
        self.frm_secondary_panel.setFrameShadow(QFrame.Raised)

        # Create 'Secondary Panel' layout
        self.lyt_secondary_panel = QVBoxLayout(self.frm_secondary_panel)

        # Configure 'Secondary Panel' header
        self.frm_secondary_panel_header = QFrame(self.frm_secondary_panel)
        self.frm_secondary_panel_header.setMinimumSize(QSize(0, 25))
        self.frm_secondary_panel_header.setFrameShape(QFrame.NoFrame)
        self.frm_secondary_panel_header.setFrameShadow(QFrame.Raised)

        # Create title for 'Secondary Panel'
        self.lbl_secondary_panel_title = QLabel(self.frm_secondary_panel_header)

        # Configure 'Secondary Panel' header layout
        self.lyt_secondary_panel_header = QHBoxLayout(self.frm_secondary_panel_header)
        self.lyt_secondary_panel_header.addWidget(self.lbl_secondary_panel_title)
        self.lyt_secondary_panel_header.setSpacing(0)
        self.lyt_secondary_panel_header.setContentsMargins(9, 0, 9, 0)

        # Configure 'Secondary Panel' stacked widget (container for 'Secondary Panel' widgets)
        self.swgt_secondary_panel = QStackedWidget(self.frm_secondary_panel)
        spol_pkg_secondary_panel = QSizePolicy(
            QSizePolicy.Ignored, QSizePolicy.Preferred
        )
        spol_pkg_secondary_panel.setHorizontalStretch(0)
        spol_pkg_secondary_panel.setVerticalStretch(0)
        self.swgt_secondary_panel.setSizePolicy(spol_pkg_secondary_panel)
        self.swgt_secondary_panel.setMinimumSize(QSize(MINIMUM_PANEL_WIDTH, 0))

        # Configure 'Secondary Panel' layout
        self.lyt_secondary_panel.addWidget(self.frm_secondary_panel_header)
        self.lyt_secondary_panel.addWidget(self.swgt_secondary_panel)
        self.lyt_secondary_panel.setSpacing(0)
        self.lyt_secondary_panel.setContentsMargins(0, 0, 0, 0)

        # Configure 'Pages' stacked widget (container for 'Page' widgets)
        self.swgt_pages = QStackedWidget(self.spl_horizontal)
        spol_pages = QSizePolicy(QSizePolicy.Ignored, QSizePolicy.Preferred)
        spol_pages.setHorizontalStretch(1)
        spol_pages.setVerticalStretch(0)
        self.swgt_pages.setSizePolicy(spol_pages)
        self.swgt_pages.setMinimumSize(QSize(300, 200))

        # Configure splitter for 'Application' page and panels
        self.spl_horizontal.addWidget(self.frm_primary_panel)
        self.spl_horizontal.addWidget(self.swgt_pages)
        self.spl_horizontal.addWidget(self.frm_secondary_panel)
        self.spl_horizontal.setOrientation(Qt.Horizontal)
        self.spl_horizontal.setHandleWidth(0)

        # Configure 'Application' layout
        self.lyt_application.addWidget(self.frm_static_tools, 0, 0, 1, 1)
        self.lyt_application.addWidget(self.spl_horizontal, 0, 1, 1, 1)
        self.lyt_application.setSpacing(0)
        self.lyt_application.setContentsMargins(0, 0, 0, 0)

        # Create 'Menu Bar'
        self.mnb_amulet = QMenuBar(self)
        self.mnb_amulet.setGeometry(QRect(0, 0, 720, 25))
        self.mnb_amulet.setMinimumSize(QSize(0, 25))

        # Create menus
        self.mn_file = QMenu(self.mnb_amulet)
        self.mn_preferences = QMenu(self.mn_file)
        self.mn_open_recent = QMenu(self.mn_file)
        self.mn_edit = QMenu(self.mnb_amulet)
        self.mn_view = QMenu(self.mnb_amulet)
        self.mn_appearance = QMenu(self.mn_view)
        self.mn_editor = QMenu(self.mn_view)

        # Configure 'Open Recent' menu
        self.act_none = QAction(self.mn_open_recent)
        self.act_clear_recently_opened = QAction(self.mn_open_recent)
        self.act_none.setEnabled(False)

        self.mn_open_recent.addAction(self.act_none)
        self.mn_open_recent.addSeparator()
        self.mn_open_recent.addAction(self.act_clear_recently_opened)

        # Configure 'Preferences' menu
        self.act_settings = QAction(self.mn_preferences)
        self.act_keyboard_shortcuts = QAction(self.mn_preferences)

        self.mn_preferences.addAction(self.act_settings)
        self.mn_preferences.addAction(self.act_keyboard_shortcuts)

        # Configure 'File' menu
        self.act_new_window = QAction(self.mn_file)
        self.act_open_project = QAction(self.mn_file)
        self.act_save = QAction(self.mn_file)
        self.act_save_as = QAction(self.mn_file)
        self.act_save_all = QAction(self.mn_file)
        self.act_close_project = QAction(self.mn_file)
        self.act_close_project.setEnabled(False)

        self.mn_file.addAction(self.act_new_window)
        self.mn_file.addSeparator()
        self.mn_file.addAction(self.act_open_project)
        self.mn_file.addAction(self.mn_open_recent.menuAction())
        self.mn_file.addSeparator()
        self.mn_file.addAction(self.act_save)
        self.mn_file.addAction(self.act_save_as)
        self.mn_file.addAction(self.act_save_all)
        self.mn_file.addSeparator()
        self.mn_file.addAction(self.mn_preferences.menuAction())
        self.mn_file.addSeparator()
        self.mn_file.addAction(self.act_close_project)

        # Configure 'Edit' menu
        self.act_undo = QAction(self.mn_edit)
        self.act_redo = QAction(self.mn_edit)
        self.act_cut = QAction(self.mn_edit)
        self.act_copy = QAction(self.mn_edit)
        self.act_paste = QAction(self.mn_edit)
        self.act_new_project = QAction(self.mn_edit)

        self.mn_edit.addAction(self.act_undo)
        self.mn_edit.addAction(self.act_redo)
        self.mn_edit.addSeparator()
        self.mn_edit.addAction(self.act_cut)
        self.mn_edit.addAction(self.act_copy)
        self.mn_edit.addAction(self.act_paste)

        # Configure 'View' menu
        self.act_show_primary_panel = QAction(self.mn_editor)
        self.act_show_secondary_panel = QAction(self.mn_editor)

        self.mn_view.addAction(self.mn_appearance.menuAction())
        self.mn_view.addAction(self.mn_editor.menuAction())
        self.mn_editor.addAction(self.act_show_primary_panel)
        self.mn_editor.addAction(self.act_show_secondary_panel)

        # Configure 'Menu Bar'
        self.mnb_amulet.addAction(self.mn_file.menuAction())
        self.mnb_amulet.addAction(self.mn_edit.menuAction())
        self.mnb_amulet.addAction(self.mn_view.menuAction())

        self.spl_horizontal.setCollapsible(1, False)
        self.spl_horizontal.setSizes([DEFAULT_PANEL_WIDTH])

        # Create 'Status Bar'
        self.sbar_amulet_status = QStatusBar(self)

        # Configure style properties
        self.frm_static_tools.setProperty("border", "none")
        self.frm_static_tools.setProperty("backgroundColor", "surface")

        self.frm_primary_panel_header.setProperty("backgroundColor", "primary")
        self.frm_primary_panel_header.setProperty("borderBottom", "surface")
        self.frm_primary_panel_header.setProperty("color", "on_surface")

        self.frm_primary_panel.setProperty("backgroundColor", "primary")
        self.frm_primary_panel.setProperty("borderRight", "surface")
        self.frm_primary_panel.setProperty("color", "on_primary")

        self.frm_secondary_panel_header.setProperty("backgroundColor", "primary")
        self.frm_secondary_panel_header.setProperty("borderBottom", "surface")
        self.frm_secondary_panel_header.setProperty("color", "on_surface")

        self.frm_secondary_panel.setProperty("backgroundColor", "primary")
        self.frm_secondary_panel.setProperty("borderLeft", "surface")
        self.frm_secondary_panel.setProperty("color", "on_primary")

        # Configure 'Application' window
        self.resize(720, 480)
        self.setCentralWidget(self.wgt_application)
        self.setMenuBar(self.mnb_amulet)
        self.setMinimumSize(QSize(720, 480))
        self.setObjectName("AmuletWindow")
        self.setStatusBar(self.sbar_amulet_status)

        # Translate widget text
        self.retranslateUi()

    def retranslateUi(self):
        # Disable formatting to condense tranlate functions
        # fmt: off
        self.setWindowTitle(QCoreApplication.translate("AmuletWindow", "Amulet Editor", None))
        self.act_open_project.setText(QCoreApplication.translate("AmuletWindow", "Open Project...", None))
        self.act_open_project.setShortcut(QCoreApplication.translate("AmuletWindow", "Ctrl+O", None))
        self.act_save.setText(QCoreApplication.translate("AmuletWindow", "Save", None))
        self.act_save.setShortcut(QCoreApplication.translate("AmuletWindow", "Ctrl+S", None))
        self.act_save_as.setText(QCoreApplication.translate("AmuletWindow", "Save As...", None))
        self.act_save_as.setShortcut(QCoreApplication.translate("AmuletWindow", "Ctrl+Shift+S", None))
        self.act_save_all.setText(QCoreApplication.translate("AmuletWindow", "Save All", None))
        self.act_settings.setText(QCoreApplication.translate("AmuletWindow", "Settings", None))
        self.act_settings.setShortcut(QCoreApplication.translate("AmuletWindow", "Ctrl+,", None))
        self.act_keyboard_shortcuts.setText(QCoreApplication.translate("AmuletWindow", "Keyboard Shortcuts", None))
        self.act_keyboard_shortcuts.setShortcut(QCoreApplication.translate("AmuletWindow", "Ctrl+K, Ctrl+S", None))
        self.act_close_project.setText(QCoreApplication.translate("AmuletWindow", "Close Project", None))
        self.act_undo.setText(QCoreApplication.translate("AmuletWindow", "Undo", None))
        self.act_undo.setShortcut(QCoreApplication.translate("AmuletWindow", "Ctrl+Z", None))
        self.act_redo.setText(QCoreApplication.translate("AmuletWindow", "Redo", None))
        self.act_redo.setShortcut(QCoreApplication.translate("AmuletWindow", "Ctrl+Y", None))
        self.act_cut.setText(QCoreApplication.translate("AmuletWindow", "Cut", None))
        self.act_cut.setShortcut(QCoreApplication.translate("AmuletWindow", "Ctrl+X", None))
        self.act_copy.setText(QCoreApplication.translate("AmuletWindow", "Copy", None))
        self.act_copy.setShortcut(QCoreApplication.translate("AmuletWindow", "Ctrl+C", None))
        self.act_paste.setText(QCoreApplication.translate("AmuletWindow", "Paste", None))
        self.act_paste.setShortcut(QCoreApplication.translate("AmuletWindow", "Ctrl+V", None))
        self.act_none.setText(QCoreApplication.translate("AmuletWindow", "None", None))
        self.act_clear_recently_opened.setText(QCoreApplication.translate("AmuletWindow", "Clear Recently Opened", None))
        self.act_new_window.setText(QCoreApplication.translate("AmuletWindow", "New Window", None))
        self.act_new_window.setShortcut(QCoreApplication.translate("AmuletWindow", "Ctrl+Shift+N", None))
        self.act_new_project.setText(QCoreApplication.translate("AmuletWindow", "New Project...", None))
        self.act_new_project.setShortcut(QCoreApplication.translate("AmuletWindow", "Ctrl+N", None))
        self.mn_file.setTitle(QCoreApplication.translate("AmuletWindow", "File", None))
        self.mn_preferences.setTitle(QCoreApplication.translate("AmuletWindow", "Preferences", None))
        self.mn_open_recent.setTitle(QCoreApplication.translate("AmuletWindow", "Open Recent", None))
        self.mn_edit.setTitle(QCoreApplication.translate("AmuletWindow", "Edit", None))
        self.mn_view.setTitle(QCoreApplication.translate("AmuletWindow", "View", None))
        self.mn_appearance.setTitle(QCoreApplication.translate("AmuletWindow", "Appearance", None))
        self.mn_editor.setTitle(QCoreApplication.translate("AmuletWindow", "Editor", None))
        self.act_show_primary_panel.setText(QCoreApplication.translate("AmuletWindow", "Show Primary Panel", None))
        self.act_show_primary_panel.setShortcut(QCoreApplication.translate("AmuletWindow", "Ctrl+B", None))
        self.act_show_secondary_panel.setText(QCoreApplication.translate("AmuletWindow", "Show Secondary Panel", None))
        self.act_show_secondary_panel.setShortcut(QCoreApplication.translate("AmuletWindow", "Ctrl+Shift+B", None))
        self.lbl_primary_panel_title.setText(QCoreApplication.translate("AmuletWindow", "App Name", None))
        # fmt: on



from datetime import datetime, timezone
from typing import Optional, Union

import amulet
from amulet.level.formats.anvil_world.format import AnvilFormat
from amulet.level.formats.leveldb_world.format import LevelDBFormat
from amulet_editor.models.text import Motd


class LevelData:
    def __init__(self, level_format: Union[AnvilFormat, LevelDBFormat]):
        self.load_format(level_format)

    def load_format(self, level_format: Union[AnvilFormat, LevelDBFormat]):
        if isinstance(level_format, AnvilFormat):
            last_played = datetime.utcfromtimestamp(level_format.last_played / 1000)
        elif isinstance(level_format, LevelDBFormat):
            last_played = datetime.utcfromtimestamp(level_format.last_played)
        else:
            raise TypeError(
                "expected AnvilFormat or LevelDBFormat object, {} found".format(
                    type(level_format)
                )
            )

        self._edition = level_format.platform.title()
        self._icon_path = (
            level_format.world_image_path
            if amulet.IMG_DIRECTORY not in level_format.world_image_path
            else None
        )
        self._name = Motd(level_format.level_name)
        self._path = level_format.path
        self._last_played = last_played.replace(tzinfo=timezone.utc)
        self._version = level_format.game_version_string.split(" ")[1]

    @property
    def edition(self) -> str:
        return self._edition

    @property
    def icon_path(self) -> Optional[str]:
        return self._icon_path

    @property
    def last_played(self) -> datetime:
        return self._last_played

    @property
    def name(self) -> Motd:
        return self._name

    @property
    def path(self) -> str:
        return self._path

    @property
    def version(self) -> str:
        return self._version



from typing import Optional
from weakref import ref

from PySide6.QtWidgets import QTreeWidgetItem, QApplication, QWidget
from PySide6.QtCore import QObject, QRect, QEvent, QPoint
from PySide6.QtGui import QMouseEvent, QPainter, QColor, QIcon

from amulet_editor.models.widgets.traceback_dialog import DisplayException
from amulet_editor.data.build import get_resource

from ._inspector import Ui_InspectionTool


class TreeWidgetItem(QTreeWidgetItem):
    @classmethod
    def create(cls, widget: QObject) -> Optional[QTreeWidgetItem]:
        if isinstance(widget, InspectorTool):
            return None
        else:
            return cls(widget)

    def __init__(self, widget: QObject):
        super().__init__([str(widget)])
        self.widget = ref(widget)
        for child in widget.children():
            item = self.create(child)
            self.addChild(item)


class CustomDraw(QObject):
    def __init__(self, target: QWidget):
        super().__init__(target)
        self.target = target
        self.background = target.grab()
        self.target.installEventFilter(self)

    def eventFilter(self, obj, event):
        if obj == self.target and event.type() == QEvent.Type.Paint:
            painter = QPainter(self.target)
            painter.drawPixmap(QPoint(0, 0), self.background)
            painter.fillRect(
                QRect(0, 0, painter.device().width(), painter.device().height()),
                QColor(115, 215, 255, 128),
            )
            painter.end()
            return True
        return super().eventFilter(obj, event)


class InspectorTool(Ui_InspectionTool):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._inspect = False
        self._highlight: Optional[tuple[QWidget, QObject]] = None

        self.inspect_button.setIcon(QIcon(get_resource("icons/tabler/click.svg")))
        self.reload_button.setIcon(QIcon(get_resource("icons/tabler/refresh.svg")))

        self.inspect_button.clicked.connect(self.inspect)
        self.reload_button.clicked.connect(self.reload)
        self.run_button.clicked.connect(self.run_code)
        self.reload()

    def reload(self):
        self.tree_widget.clear()
        for window in QApplication.topLevelWidgets():
            root = TreeWidgetItem.create(window)
            if root is not None:
                self.tree_widget.addTopLevelItem(root)
        if self.tree_widget.topLevelItemCount():
            self.tree_widget.topLevelItem(0)

    def inspect(self):
        self.setMouseTracking(True)
        self.grabMouse()
        self._inspect = True

    def _remove_highlight(self):
        if self._highlight is not None:
            widget, filt = self._highlight
            widget.removeEventFilter(filt)
            widget.update()
            self._highlight = None

    def mouseMoveEvent(self, event: QMouseEvent):
        if self._inspect:
            widget = QApplication.widgetAt(event.globalPosition().toPoint())
            if not widget:
                return
            if widget.topLevelWidget() is self:
                return
            if self._highlight is not None:
                if self._highlight[0] is widget:
                    return
                else:
                    self._remove_highlight()

            filt = CustomDraw(widget)
            widget.update()
            self._highlight = widget, filt
        else:
            super().mouseMoveEvent(event)

    def mousePressEvent(self, event: QMouseEvent):
        if self._inspect:
            self._inspect = False
            self.releaseMouse()
            self.setMouseTracking(False)
            self._remove_highlight()
        else:
            super().mousePressEvent(event)

    def run_code(self):
        item = self.tree_widget.currentItem()
        if not isinstance(item, TreeWidgetItem):
            return
        obj = item.widget()
        if obj is None:
            print("Selected object no longer exists.")
        else:
            with DisplayException("Error running user code."):
                eval(self.code_editor.toPlainText(), {}, {"obj": obj})


_inspector = None


def show_inspector():
    global _inspector
    if _inspector is None:
        _inspector = InspectorTool()
        _inspector.show()
    _inspector.raise_()



from PySide6.QtGui import QMatrix4x4


class Drawable:
    def paintGL(self, projection_matrix: QMatrix4x4, view_matrix: QMatrix4x4):
        raise NotImplementedError

    def initializeGL(self):
        """Initialise the OpenGL data."""
        raise NotImplementedError



"""Utility class to create horizontal line separators."""

from typing import Optional
from PySide6.QtWidgets import QFrame, QWidget


class QHLine(QFrame):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.setFrameShape(QFrame.Shape.HLine)
        self.setProperty("borderTop", "surface")



from __future__ import annotations

from typing import Optional

from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import QWidget, QVBoxLayout

from .home import HomePage
from .open_world import OpenWorldPage

from amulet_team_main_window.application.windows.main_window import View
from amulet_team_locale import set_locale


class HomeView(QWidget, View):
    def __init__(
        self, parent: Optional[QWidget] = None, f: Qt.WindowType = Qt.WindowType.Widget
    ):
        super().__init__(parent, f)
        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._set_landing_page()

    def activate_view(self):
        self._set_landing_page()

    def setCentralWidget(self, widget: QWidget):
        for _ in range(self._layout.count()):
            old_widget = self._layout.takeAt(0)
            if old_widget is not None:
                old_widget.widget().deleteLater()
        self._layout.addWidget(widget)

    def _set_landing_page(self):
        page = HomePage(self)
        # Connect signals
        page.btn_open_world.clicked.connect(self._set_open_world_page)
        # page.crd_new_project.clicked.connect(
        #     partial(self.set_menu_page, NewProjectMenu)
        # )

        @Slot(int)
        def _locale_change(index: int):
            set_locale(page.cbo_language.currentData())

        page.cbo_language.currentIndexChanged.connect(_locale_change)
        self.setCentralWidget(page)

    def _set_open_world_page(self):
        page = OpenWorldPage()
        page.btn_back.clicked.connect(self._set_landing_page)
        self.setCentralWidget(page)



from amulet_editor.models.package import AmuletTool, AmuletView
from amulet_editor.tools.programs.pages._programs import ProgramsPage
from amulet_editor.tools.programs.panels._programs import ProgramsPanel
from PySide6.QtCore import QObject
from PySide6.QtWidgets import QWidget


class Programs(AmuletTool):
    def __init__(self):
        self._page = AmuletView()
        self._primary_panel = AmuletView()
        self._secondary_panel = AmuletView()

        self.view_manager = ProgramsManager(self)

    def set_page(self, widget: QWidget):
        self._page.setWidget(widget)

    def set_primary_panel(self, widget: QWidget):
        self._primary_panel.setWidget(widget)

    def set_secondary_panel(self, widget: QWidget):
        self._secondary_panel.setWidget(widget)

    @property
    def page(self) -> AmuletView:
        return self._page

    @property
    def primary_panel(self) -> AmuletView:
        return self._primary_panel

    @property
    def secondary_panel(self) -> AmuletView:
        return self._secondary_panel

    @property
    def name(self) -> str:
        return "Programs"

    @property
    def icon_name(self) -> str:
        return "app-window.svg"


class ProgramsManager(QObject):
    def __init__(self, plugin: Programs):
        super().__init__()

        self.plugin = plugin

        self.programs_page = ProgramsPage()

        self.programs_panel = ProgramsPanel()

        self.set_programs_page()
        self.set_programs_panel()

    def set_programs_page(self):
        page = self.programs_page

        self.plugin.set_page(page)

    def set_programs_panel(self):
        panel = self.programs_panel

        self.plugin.set_primary_panel(panel)



import logging
import traceback as tb
from amulet_editor.application._invoke import invoke
from ._cls import _AmuletTracebackDialog


def display_exception_blocking(title: str = "", error: str = "", traceback: str = ""):
    """
    Display an exception window.
    This must be called from the main thread.
    This blocks until the user closes the window.

    :param title: The title of the dialog.
    :param error: A user-readable description of the error context.
    :param traceback: The traceback to display in the dialog.
    """
    dialog = _AmuletTracebackDialog(title=title, error=error, traceback=traceback)
    dialog.exec()


def display_exception(title: str = "", error: str = "", traceback: str = ""):
    """
    Display an exception window.
    This is processed when control returns to the main thread.
    This is thread safe.

    :param title: The title of the dialog.
    :param error: A user-readable description of the error context.
    :param traceback: The traceback to display in the dialog.
    """
    invoke(
        lambda: display_exception_blocking(
            title=title, error=error, traceback=traceback
        )
    )


class DisplayException:
    """
    A context manager class to display the traceback dialog when an error occurs.
    It will also log the exception to the logging module and optionally suppress the exception.
    """

    def __init__(self, msg: str, *, suppress=False, log=logging.getLogger()):
        self._msg = msg
        self._suppress = suppress
        self._log = log

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type and isinstance(exc_val, Exception):
            self._log.exception(exc_val)
            display_exception(
                title=self._msg,
                error=str(exc_val),
                traceback="".join(tb.format_tb(exc_tb)),
            )
            return self._suppress
        return False



import argparse
import logging
import subprocess
import sys
from typing import Protocol, Optional
from amulet_editor.data.paths._application import (
    DefaultDataDir,
    DefaultConfigDir,
    DefaultCacheDir,
    DefaultLogDir,
)


BROKER = "BROKER"


class Args(Protocol):
    level_path: Optional[str]
    data_dir: Optional[str]
    config_dir: Optional[str]
    cache_dir: Optional[str]
    log_dir: Optional[str]
    logging_level: int
    logging_format: str
    trace: bool


_args: Optional[Args] = None


def parse_args() -> Args:
    global _args

    if _args is None:
        parser = argparse.ArgumentParser(
            prog="Amulet Editor",
            description="Amulet is a Minecraft world editing application.",
            epilog="The following arguments are a subset of the full arguments.",
        )

        parser.add_argument(
            "--level_path",
            type=str,
            help="The Minecraft world or structure to open. Default opens no level",
            action="store",
            dest="level_path",
            default=None,
        )

        parser.add_argument(
            "--data_dir",
            type=str,
            help=f"The directory to store application data in. Default is {DefaultDataDir}",
            action="store",
            dest="data_dir",
            default=None,
        )

        parser.add_argument(
            "--config_dir",
            type=str,
            help=f"The directory to store application config files in. Default is {DefaultConfigDir}",
            action="store",
            dest="config_dir",
            default=None,
        )

        parser.add_argument(
            "--cache_dir",
            type=str,
            help=f"The directory to store cache files in. Default is {DefaultCacheDir}",
            action="store",
            dest="cache_dir",
            default=None,
        )

        parser.add_argument(
            "--log_dir",
            type=str,
            help=f"The directory to store log files in. Default is {DefaultLogDir}",
            action="store",
            dest="log_dir",
            default=None,
        )

        parser.add_argument(
            "--logging_level",
            type=int,
            help="The logging level to set. CRITICAL=50, ERROR=40, WARNING=30, INFO=20, DEBUG=10. Default is WARNING",
            action="store",
            dest="logging_level",
            default=logging.WARNING,
        )

        parser.add_argument(
            "--logging_format",
            type=str,
            help='The logging format to use. Default is "%(levelname)s - %(message)s"',
            action="store",
            dest="logging_format",
            default="%(levelname)s - %(message)s",
        )

        parser.add_argument(
            "--trace",
            help="If defined, print the qualified name of each function as it is called. Useful if the program crashes due to an access violation and you don't know where it came from.",
            action="store_true",
            dest="trace",
        )

        _args, _ = parser.parse_known_args()

    return _args  # noqa


# TODO: move this somewhere more sensible
def spawn_process(path: str = None):
    """Spawn the broker process passing over the input CLI values."""
    this_args = parse_args()
    new_args = [sys.executable, sys.argv[0]]
    if path is not None:
        new_args += ["--level_path", path]
    new_args += [
        "--logging_level",
        str(this_args.logging_level),
        "--logging_format",
        this_args.logging_format,
    ]
    if this_args.trace:
        new_args.append("--trace")
    subprocess.Popen(
        new_args,
        start_new_session=True,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )



from typing import Optional
import os
from PySide6.QtCore import QTranslator, QLocale, QDir


class ATranslator(QTranslator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._translations: dict[str, str] = {}
        self._path = ""

    def filePath(self) -> str:
        return super().filePath() or self._path

    def isEmpty(self) -> bool:
        return super().isEmpty() and not self._translations

    def load_lang(
        self,
        locale: QLocale,
        filename: str,
        prefix: str = "",
        directory: str = "",
        suffix: str = ".lang",
    ) -> bool:
        self._translations.clear()

        def get_codes():
            for language_ in locale.uiLanguages():
                language_split = language_.split("-")
                for i in range(len(language_split), -1, -1):
                    yield "_".join(language_split[:i])
            yield "en"

        directory = directory or QDir.currentPath()

        for ui_language_name in get_codes():
            for ext in [suffix, ""]:
                path = (
                    directory + os.path.sep + filename + prefix + ui_language_name + ext
                )
                if os.path.isfile(path):
                    with open(path) as lang:
                        for line in lang.readlines():
                            line = line.strip()
                            if not line or line[0] == "#":
                                continue
                            else:
                                line_split = line.split("=", 1)
                                if len(line_split) != 2:
                                    # TODO: error dialog
                                    continue
                                self._translations[line_split[0]] = line_split[1]
                    return True
        return False

    def translate(
        self,
        context: str,
        source_text: str,
        disambiguation: Optional[bytes] = None,
        n: int = -1,
    ) -> str:
        return super().translate(
            context, source_text, disambiguation, n
        ) or self._translations.get(f"{context}.{source_text}", None)



import glob
from PyInstaller.utils.hooks import (
    collect_data_files,
    collect_submodules,
)

hiddenimports = collect_submodules("PySide6")
print("added qt libraries")
print([(path, ".") for path in glob.glob("Qt/lib/*.so*")])
binaries = [(path, ".") for path in glob.glob("Qt/lib/*.so*")]
datas = collect_data_files("PySide6", includes=["*.pyi", "py.typed"])



from __future__ import annotations
import logging
from typing import Optional, Generator, Callable, Iterator
import ctypes
from threading import Lock
from weakref import WeakKeyDictionary, WeakValueDictionary, ref
from collections.abc import MutableMapping
import bisect
import numpy

from PySide6.QtCore import QThread, Signal, QObject, Slot, QTimer, QCoreApplication
from PySide6.QtGui import QMatrix4x4, QOpenGLContext, QOffscreenSurface
from PySide6.QtOpenGL import (
    QOpenGLVertexArrayObject,
    QOpenGLBuffer,
    QOpenGLShaderProgram,
    QOpenGLShader,
    QOpenGLTexture,
)
from shiboken6 import VoidPtr, isValid
from OpenGL.GL import (
    GL_FLOAT,
    GL_FALSE,
    GL_TRIANGLES,
    GL_CULL_FACE,
    GL_BACK,
    GL_DEPTH_TEST,
    GL_LEQUAL,
    GL_BLEND,
    GL_SRC_ALPHA,
    GL_ONE_MINUS_SRC_ALPHA,
)

from amulet.api.data_types import Dimension
from amulet.api.level import BaseLevel
from amulet.api.errors import ChunkLoadError, ChunkDoesNotExist
from amulet.api.chunk import Chunk

from ._drawable import Drawable
from ._resource_pack import (
    OpenGLResourcePack,
    get_gl_resource_pack_container,
)
from amulet_editor.application._invoke import invoke

try:
    from .chunk_builder_cy import create_lod0_chunk
except:
    raise Exception(
        "Could not import cython chunk mesher. The cython code must be compiled first."
    )

FloatSize = ctypes.sizeof(ctypes.c_float)

log = logging.getLogger(__name__)

ChunkKey = tuple[Dimension, int, int]


class SharedVBOManager(QObject):
    """
    This class manages creating and destroying VBOs so that you don't have to.
    The VBO is created in a context that shares the global share context. Creation is done on the main thread.
    All methods are thread safe.
    """

    _context: QOpenGLContext
    _surface: QOffscreenSurface
    _vbos: set[QOpenGLBuffer]

    def __init__(self):
        """Use new class method"""
        if QThread.currentThread() is not QCoreApplication.instance().thread():
            raise RuntimeError(
                "SharedVBOManager must be constructed on the main thread."
            )

        super().__init__()
        context = self._context = QOpenGLContext()
        global_context = QOpenGLContext.globalShareContext()
        if global_context is None:
            raise RuntimeError("Global OpenGL context does not exist.")
        self._context.setShareContext(global_context)
        self._context.create()

        # Create the surface
        surface = self._surface = QOffscreenSurface()
        self._surface.create()

        lock = self._lock = Lock()
        vbos = self._vbos = set()

        def destroy():
            with lock:
                if not context.makeCurrent(surface):
                    raise RuntimeError("Could not make context current.")

                for vbo in vbos:
                    vbo.destroy()
                vbos.clear()

                context.doneCurrent()

        self.destroyed.connect(destroy)

    def create_vbo(self, buffer: bytes) -> QOpenGLBuffer:
        def create_vbo():
            with self._lock:
                start_context = QOpenGLContext.currentContext()
                start_surface = None
                if start_context is not None:
                    start_surface = start_context.surface()
                    start_context.doneCurrent()

                try:
                    if not self._context.makeCurrent(self._surface):
                        raise RuntimeError("Could not make context current.")

                    vbo = QOpenGLBuffer()
                    vbo.create()
                    vbo.bind()
                    vbo.allocate(buffer, len(buffer))
                    vbo.release()
                    self._vbos.add(vbo)
                    self._context.doneCurrent()

                finally:
                    if start_context is not None and not start_context.makeCurrent(
                        start_surface
                    ):
                        raise RuntimeError("Could not make original context current.")

                return vbo

        return invoke(create_vbo)

    def destroy_vbo(self, vbo: QOpenGLBuffer):
        def destroy_vbo():
            with self._lock:
                start_context = QOpenGLContext.currentContext()
                start_surface = None
                if start_context is not None:
                    start_surface = start_context.surface()
                    start_context.doneCurrent()

                try:
                    if vbo not in self._vbos:
                        raise RuntimeError(
                            "vbo was not created by this class or has already been destroyed."
                        )

                    if not self._context.makeCurrent(self._surface):
                        raise RuntimeError("Could not make context current.")
                    vbo.destroy()
                    self._context.doneCurrent()
                    log.debug("Destroyed vbo")

                finally:
                    if start_context is not None and not start_context.makeCurrent(
                        start_surface
                    ):
                        raise RuntimeError("Could not make original context current.")

        invoke(destroy_vbo)


class SharedChunkGeometry(QObject):
    """
    A class holding all the shared chunk data.
    The vbo will be deallocated when this instance is destroyed.
    """

    vbo: QOpenGLBuffer
    vertex_count: int
    _resource_pack: OpenGLResourcePack
    texture: QOpenGLTexture

    def __init__(
        self, vbo: QOpenGLBuffer, vertex_count: int, resource_pack: OpenGLResourcePack
    ):
        super().__init__()
        self.vbo = vbo
        self.vertex_count = vertex_count
        self._resource_pack = resource_pack
        self.texture = self._resource_pack.get_texture()


class SharedChunkData(QObject):
    """
    The instances of this class belong to the :class:`SharedLevelGeometry` that created them.
    All data will be destroyed when the last reference to the instance is lost.
    """

    # Constant data
    # The world transform of the data.
    model_transform: QMatrix4x4

    # Variable data
    # The shared OpenGL data.
    # This variable can get modified, you must hold a strong reference to this object and access it from there.
    # It may be None initially until geometry_changed is emitted.
    geometry: Optional[SharedChunkGeometry]

    # Signals
    # Emitted when the geometry has been modified.
    geometry_changed = Signal()

    def __init__(self, model_transform: QMatrix4x4):
        super().__init__()
        self.model_transform: QMatrix4x4 = model_transform
        self.geometry = None


class ChunkGeneratorWorker(QObject):
    """
    This object exists in a secondary thread so that chunk generation does not block the main thread.
    The OpenGL calls need to be done on the main thread.
    """

    def __init__(self):
        super().__init__()
        self.generate_chunk.connect(self._generate_chunk)

    generate_chunk = Signal(object, object, object, object)

    @Slot(object, object, object, object)
    def _generate_chunk(
        self,
        level: BaseLevel,
        vbo_manager: SharedVBOManager,
        chunk_key: ChunkKey,
        chunk_data: SharedChunkData,
    ):
        try:
            resource_pack_container = get_gl_resource_pack_container(level)
            if not resource_pack_container.loaded:
                # The resource pack does not exist yet so push the job to the back of the queue
                QTimer.singleShot(
                    100,
                    lambda: self.generate_chunk.emit(
                        level, vbo_manager, chunk_key, chunk_data
                    ),
                )
                return

            resource_pack = resource_pack_container.resource_pack

            dimension, cx, cz = chunk_key

            buffer = b""
            try:
                chunk = level.get_chunk(cx, cz, dimension)
            except ChunkDoesNotExist:
                # TODO: Add void geometry
                pass
            except ChunkLoadError:
                # TODO: Add error geometry
                log.exception(f"Error loading chunk {chunk_key}", exc_info=True)
                pass
                # self._create_error_geometry()
            else:
                chunk_verts, chunk_verts_translucent = create_lod0_chunk(
                    resource_pack,
                    numpy.zeros(3, dtype=numpy.int32),
                    self._sub_chunks(level, dimension, cx, cz, chunk),
                    chunk.block_palette,
                )
                verts = chunk_verts + chunk_verts_translucent
                if verts:
                    buffer = numpy.concatenate(verts).tobytes()

            log.debug(f"Generated array for {chunk_key}")

            vbo = vbo_manager.create_vbo(buffer)

            vertex_count = len(buffer) // (12 * FloatSize)

            chunk_data.geometry = SharedChunkGeometry(vbo, vertex_count, resource_pack)
            # When the container gets garbage collected, destroy the vbo
            chunk_data.geometry.destroyed.connect(lambda: vbo_manager.destroy_vbo(vbo))
            chunk_data.geometry_changed.emit()

            log.debug(f"Generated chunk {chunk_key}")
        except Exception as e:
            log.exception(e)

    def _sub_chunks(
        self, level: BaseLevel, dimension: Dimension, cx: int, cz: int, chunk: Chunk
    ) -> list[tuple[numpy.ndarray, int]]:
        """
        Create sub-chunk arrays that extend into the neighbour sub-chunks by one block.

        :param dimension: The dimension the chunk came from
        :param cx: The chunk x coordinate of the chunk
        :param cz: The chunk z coordinate of the chunk
        :param chunk: The chunk object
        :return: A list of tuples containing the larger block array and the location of the sub-chunk
        """
        blocks = chunk.blocks
        sub_chunks = []
        neighbour_chunks = {}
        for dx, dz in ((-1, 0), (1, 0), (0, -1), (0, 1)):
            try:
                neighbour_chunks[(dx, dz)] = level.get_chunk(
                    cx + dx, cz + dz, dimension
                ).blocks
            except ChunkLoadError:
                continue

        for cy in blocks.sub_chunks:
            sub_chunk = blocks.get_sub_chunk(cy)
            larger_blocks = numpy.zeros(
                sub_chunk.shape + numpy.array((2, 2, 2)), sub_chunk.dtype
            )
            # if self._limit_bounds:
            #     sub_chunk_box = SelectionBox.create_sub_chunk_box(cx, cy, cz)
            #     if level.bounds(self.dimension).intersects(sub_chunk_box):
            #         boxes = level.bounds(self.dimension).intersection(
            #             sub_chunk_box
            #         )
            #         for box in boxes.selection_boxes:
            #             larger_blocks[1:-1, 1:-1, 1:-1][
            #                 box.sub_chunk_slice(self.cx, cy, self.cz)
            #             ] = sub_chunk[box.sub_chunk_slice(self.cx, cy, self.cz)]
            #     else:
            #         continue
            # else:
            #     larger_blocks[1:-1, 1:-1, 1:-1] = sub_chunk
            larger_blocks[1:-1, 1:-1, 1:-1] = sub_chunk
            for chunk_offset, neighbour_blocks in neighbour_chunks.items():
                if cy not in neighbour_blocks:
                    continue
                if chunk_offset == (-1, 0):
                    larger_blocks[0, 1:-1, 1:-1] = neighbour_blocks.get_sub_chunk(cy)[
                        -1, :, :
                    ]
                elif chunk_offset == (1, 0):
                    larger_blocks[-1, 1:-1, 1:-1] = neighbour_blocks.get_sub_chunk(cy)[
                        0, :, :
                    ]
                elif chunk_offset == (0, -1):
                    larger_blocks[1:-1, 1:-1, 0] = neighbour_blocks.get_sub_chunk(cy)[
                        :, :, -1
                    ]
                elif chunk_offset == (0, 1):
                    larger_blocks[1:-1, 1:-1, -1] = neighbour_blocks.get_sub_chunk(cy)[
                        :, :, 0
                    ]
            if cy - 1 in blocks:
                larger_blocks[1:-1, 0, 1:-1] = blocks.get_sub_chunk(cy - 1)[:, -1, :]
            if cy + 1 in blocks:
                larger_blocks[1:-1, -1, 1:-1] = blocks.get_sub_chunk(cy + 1)[:, 0, :]
            sub_chunks.append((larger_blocks, cy * 16))
        return sub_chunks


class ChunkGenerator(QObject):
    _thread: QThread
    _worker: ChunkGeneratorWorker

    def __init__(self):
        super().__init__()
        thread = self._thread = QThread()
        self._thread.start()
        self._worker = ChunkGeneratorWorker()
        self._worker.moveToThread(self._thread)

        def destroy():
            self._worker.deleteLater()
            log.debug("Waiting for chunk generation thread to finish")
            thread.quit()
            thread.wait()
            log.debug("Chunk generation thread has finished")
            log.debug("destroyed ChunkGenerator")

        self.destroyed.connect(destroy)
        QCoreApplication.instance().aboutToQuit.connect(self.deleteLater)

    def generate_chunk(
        self,
        level: BaseLevel,
        vbo_manager: SharedVBOManager,
        chunk_key: ChunkKey,
        chunk: SharedChunkData,
    ):
        """
        Async function to generate the chunk VBO.

        :param level: The level to pull chunks from.
        :param vbo_manager: The instance that will manage vbo creation.
        :param chunk_key: The key of the chunk to get.
        :param chunk: The SharedChunkData to store the value in and notify through.
        :return: Returns None immediately.
        """
        self._worker.generate_chunk.emit(level, vbo_manager, chunk_key, chunk)


class SharedLevelGeometry(QObject):
    """
    A class holding the shared level geometry relating to one level.
    This class must exist on the main thread.
    """

    # Class variables
    _instances_lock = Lock()
    _instances: WeakKeyDictionary[BaseLevel, SharedLevelGeometry] = {}

    # Instance variables
    _level: Callable[[], Optional[BaseLevel]]
    _chunks_lock: Lock
    # Store the chunk data weakly so that it gets automatically deallocated
    _chunks: WeakValueDictionary[ChunkKey, SharedChunkData]

    _chunk_generator: ChunkGenerator
    _vbo_manager: SharedVBOManager

    @classmethod
    def instance(cls, level: BaseLevel) -> SharedLevelGeometry:
        with cls._instances_lock:
            if level not in cls._instances:
                cls._instances[level] = invoke(lambda: SharedLevelGeometry(level))
            return cls._instances[level]

    def __init__(self, level: BaseLevel):
        """To get an instance of this class you should use :classmethod:`instance`"""
        super().__init__()
        self._level = ref(level)
        self._chunks_lock = Lock()
        self._chunks = WeakValueDictionary()

        self._vbo_manager = invoke(SharedVBOManager)
        self._chunk_generator = ChunkGenerator()

        self._resource_pack_container = get_gl_resource_pack_container(level)
        self._resource_pack_container.changed.connect(self._resource_pack_changed)

        QCoreApplication.instance().aboutToQuit.connect(self.deleteLater)

    def get_chunk(self, chunk_key: ChunkKey) -> SharedChunkData:
        """Get the geometry for a chunk."""
        with self._chunks_lock:
            chunk = self._chunks.get(chunk_key)
            if chunk is None:
                dimension, cx, cz = chunk_key
                transform = QMatrix4x4()
                transform.translate(cx * 16, 0, cz * 16)
                chunk = self._chunks[chunk_key] = SharedChunkData(transform)
                self._chunk_generator.generate_chunk(
                    self._level(), self._vbo_manager, chunk_key, chunk
                )
            return chunk

    def _resource_pack_changed(self):
        # The geometry of all loaded chunks needs to be rebuilt.
        with self._chunks_lock:
            for chunk_key, chunk in self._chunks.items():
                self._chunk_generator.generate_chunk(
                    self._level(), self._vbo_manager, chunk_key, chunk
                )


class WidgetChunkData(QObject):
    shared: SharedChunkData
    geometry: Optional[SharedChunkGeometry]
    vao: Optional[QOpenGLVertexArrayObject]

    geometry_changed = Signal()

    def __init__(self, shared: SharedChunkData):
        super().__init__()
        self.shared = shared
        self.geometry = None
        self.vao = None
        self.shared.geometry_changed.connect(self.geometry_changed)


def empty_iterator():
    yield from ()


def get_grid_spiral(
    dimension: Dimension, cx: int, cz: int, r: int
) -> Generator[ChunkKey, None, None]:
    """A generator that yields a 2D grid spiraling from the centre."""
    sign = 1
    length = 1
    for _ in range(r * 2 + 1):
        for _ in range(length):
            yield dimension, cx, cz
            cx += sign
        for _ in range(length):
            yield dimension, cx, cz
            cz += sign
        sign *= -1
        length += 1


class ChunkContainer(MutableMapping):
    def __init__(self):
        self._chunks: dict[ChunkKey, WidgetChunkData] = {}
        self._order: list[ChunkKey] = []
        self._x = 0
        self._z = 0

    def set_position(self, cx, cz):
        self._x = cx
        self._z = cz
        self._order = sorted(self._order, key=self._dist)

    def __contains__(self, k: ChunkKey):
        return k in self._chunks

    def _dist(self, k: ChunkKey):
        return -abs(k[1] - self._x) - abs(k[2] - self._z)

    def __setitem__(self, k: ChunkKey, v: WidgetChunkData):
        if k not in self._chunks:
            self._order.insert(
                bisect.bisect_left(self._order, self._dist(k), key=self._dist), k
            )
        self._chunks[k] = v

    def __delitem__(self, v: ChunkKey) -> None:
        del self._chunks[v]
        self._order.remove(v)

    def __getitem__(self, k: ChunkKey) -> WidgetChunkData:
        return self._chunks[k]

    def __len__(self) -> int:
        return len(self._chunks)

    def __iter__(self) -> Iterator[ChunkKey]:
        yield from self._order


class WidgetLevelGeometry(QObject, Drawable):
    """
    A class holding the level geometry data relating to one widget.
    This holds all non-shared data and references to the shared data.
    When this object is deleted it will automatically destroy its own data.

    The methods in this class must only be called from the thread the owner QOpenGLWidget is in.
    """

    _shared: SharedLevelGeometry

    _load_radius: int
    _unload_radius: int
    _dimension: Optional[Dimension]
    _camera_chunk: Optional[tuple[int, int]]
    _chunk_finder: Generator[ChunkKey, None, None]
    _running: bool
    _generation_count: int

    # OpenGL attributes
    _context: Optional[QOpenGLContext]
    _surface: QOffscreenSurface
    _program: Optional[QOpenGLShaderProgram]
    _matrix_location: Optional[int]
    _texture_location: Optional[int]
    _chunks: ChunkContainer
    _pending_chunks: dict[ChunkKey, WidgetChunkData]

    # The geometry has changed and needs repainting.
    geometry_changed = Signal()

    def __init__(self, level: BaseLevel):
        super().__init__()
        self._shared = SharedLevelGeometry.instance(level)

        self._load_radius = 5
        self._unload_radius = 10
        self._dimension = None
        self._camera_chunk = None
        self._chunk_finder = empty_iterator()
        self._running = False
        self._generation_count = 0

        self._context = None
        self._surface = QOffscreenSurface()
        self._surface.create()
        self._program = None
        self._matrix_location = None
        self._texture_location = None
        self._chunks = ChunkContainer()
        self._pending_chunks = {}

        self._queue_chunk.connect(self._process_chunk)

    def initializeGL(self):
        """
        Initialise the opengl state.
        The widget context must be current before calling this.
        This must only be called from the QOpenGLWidget that this instance is associated with.
        """
        context = QOpenGLContext.currentContext()
        if not QOpenGLContext.areSharing(context, QOpenGLContext.globalShareContext()):
            raise RuntimeError(
                "The widget context is not sharing with the global context."
            )

        # Make sure any old data no longer exists.
        self._destroy_gl()
        self._context = context
        self._context.aboutToBeDestroyed.connect(self._destroy_gl)

        # Initialise the shader
        self._program = QOpenGLShaderProgram()
        self._program.addShaderFromSourceCode(
            QOpenGLShader.ShaderTypeBit.Vertex,
            """#version 150
            in vec3 position;
            in vec2 vTexCoord;
            in vec4 vTexOffset;
            in vec3 vTint;

            out vec2 fTexCoord;
            out vec4 fTexOffset;
            out vec3 fTint;

            uniform mat4 transformation_matrix;

            void main() {
                gl_Position = transformation_matrix * vec4(position, 1.0);
                fTexCoord = vTexCoord;
                fTexOffset = vTexOffset;
                fTint = vTint;
            }""",
        )

        self._program.addShaderFromSourceCode(
            QOpenGLShader.ShaderTypeBit.Fragment,
            """#version 150
            in vec2 fTexCoord;
            in vec4 fTexOffset;
            in vec3 fTint;

            out vec4 outColor;

            uniform sampler2D image;

            void main(){
                vec4 texColor = texture(
                    image,
                    vec2(
                        mix(fTexOffset.x, fTexOffset.z, mod(fTexCoord.x, 1.0)),
                        mix(fTexOffset.y, fTexOffset.w, mod(fTexCoord.y, 1.0))
                    )
                );
                if(texColor.a < 0.02)
                    discard;
                texColor.xyz = texColor.xyz * fTint * 0.85;
                outColor = texColor;
            }""",
        )

        self._program.bindAttributeLocation("position", 0)
        self._program.bindAttributeLocation("vTexCoord", 1)
        self._program.bindAttributeLocation("vTexOffset", 2)
        self._program.bindAttributeLocation("vTint", 3)
        self._program.link()
        self._program.bind()
        self._matrix_location = self._program.uniformLocation("transformation_matrix")
        self._texture_location = self._program.uniformLocation("image")
        self._program.release()

    def _destroy_gl(self):
        if (
            self._context is not None
            and isValid(self._context)
            and self._context.isValid()
        ):
            if QOpenGLContext.currentContext() is not self._context:
                # Enable the context if it isn't.
                # Use an offscreen surface because the widget surface may no longer exist.
                if not self._context.makeCurrent(self._surface):
                    raise RuntimeError("Could not make context current.")

            self._clear_chunks_no_context()

        self._program = None
        self._matrix_location = None
        self._texture_location = None
        self._context = None

    def __del__(self):
        self._destroy_gl()
        log.debug("deleted WidgetLevelGeometry")

    def paintGL(self, projection_matrix: QMatrix4x4, view_matrix: QMatrix4x4):
        """
        Draw the level.
        The context must be active before calling this.

        :param projection_matrix: The camera internal projection matrix.
        :param view_matrix: The camera external matrix.
        """
        context = QOpenGLContext.currentContext()
        if context is None or not QOpenGLContext.areSharing(context, self._context):
            raise RuntimeError("Context is not valid")

        f = QOpenGLContext.currentContext().functions()
        f.glEnable(GL_DEPTH_TEST)
        f.glDepthFunc(GL_LEQUAL)
        f.glEnable(GL_BLEND)
        f.glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        f.glEnable(GL_CULL_FACE)
        f.glCullFace(GL_BACK)

        # Draw the geometry
        self._program.bind()

        # Init the texture
        self._program.setUniformValue1i(self._texture_location, 0)

        transform = projection_matrix * view_matrix
        for chunk_data in self._chunks.values():
            self._program.setUniformValue(
                self._matrix_location, transform * chunk_data.shared.model_transform
            )
            chunk_data.geometry.texture.bind(0)
            chunk_data.vao.bind()
            f.glDrawArrays(GL_TRIANGLES, 0, chunk_data.geometry.vertex_count)
            chunk_data.vao.release()

        self._program.release()

    def start(self):
        """Start chunk generation"""
        self._running = True
        self._update_chunk_finder()

    def stop(self):
        self._running = False

    def _update_chunk_finder(self):
        if self._dimension is None or self._camera_chunk is None:
            empty_iterator()
        else:
            cx, cz = self._camera_chunk
            self._chunk_finder = get_grid_spiral(
                self._dimension, cx, cz, self._load_radius
            )
            self._queue_next_chunk()

    def _clear_chunks(self):
        """Unload all chunk data. It is the job of the caller to handle the chunk generator."""
        if self._context is None:
            return
        if not self._context.makeCurrent(self._surface):
            raise RuntimeError("Could not make context current.")
        self._clear_chunks_no_context()
        self._context.doneCurrent()

    def _clear_chunks_no_context(self):
        """Unload all chunk data. Caller manages the GL context."""
        for chunk in self._chunks.values():
            if isValid(chunk.vao):
                # I don't understand where this is being destroyed from
                chunk.vao.destroy()
        self._context.doneCurrent()
        self._chunks.clear()
        self._pending_chunks.clear()

    def _clear_far_chunks(self):
        """
        Unload all chunk data outside the unload render distance.
        It is the job of the caller to handle the chunk generator.
        """
        if self._context is None:
            return
        if self._camera_chunk is None:
            return

        if not self._context.makeCurrent(self._surface):
            raise RuntimeError("Could not make context current.")

        camera_cx, camera_cz = self._camera_chunk
        min_cx = camera_cx - self._unload_radius
        max_cx = camera_cx + self._unload_radius
        min_cz = camera_cz - self._unload_radius
        max_cz = camera_cz + self._unload_radius

        for container in (self._chunks, self._pending_chunks):
            remove_chunk_keys = []
            for chunk_key, chunk in container.items():
                _, chunk_cx, chunk_cz = chunk_key
                if (
                    chunk_cx < min_cx
                    or chunk_cx > max_cx
                    or chunk_cz < min_cz
                    or chunk_cz > max_cz
                ):
                    if chunk.vao is not None:
                        chunk.vao.destroy()
                    remove_chunk_keys.append(chunk_key)
            for chunk_key in remove_chunk_keys:
                del container[chunk_key]

        self._context.doneCurrent()

    def set_dimension(self, dimension: Dimension):
        if dimension != self._dimension:
            self._dimension = dimension
            self._clear_chunks()
            self._update_chunk_finder()

    def set_location(self, cx: int, cz: int):
        """Set the chunk the camera is in."""
        cx = int(cx)
        cz = int(cz)
        location = (cx, cz)
        if location != self._camera_chunk:
            self._camera_chunk = location
            self._clear_far_chunks()
            self._update_chunk_finder()
            self._chunks.set_position(cx, cz)

    def set_render_distance(self, load_distance: int, unload_distance: int):
        """
        Set the render distance attributes.

        :param load_distance: The radius within which to load chunks.
        :param unload_distance: The radius outside which chunks will be unloaded.
        """
        if unload_distance < load_distance + 2:
            # Make sure that unload distance is larger than load distance plus a margin
            unload_distance = load_distance + 2

        if unload_distance < self._unload_radius:
            # TODO: Unload chunks outside the radius
            raise NotImplementedError

        self._load_radius = load_distance
        self._unload_radius = unload_distance
        self._clear_far_chunks()
        self._update_chunk_finder()

    def _queue_next_chunk(self):
        if self._running and self._generation_count == 0:
            # The instance is running and there are no existing generation calls running.
            self._generation_count += 1
            self._queue_chunk.emit()

    _queue_chunk = Signal()

    def _process_chunk(self):
        """Generate a chunk. This must not be called directly."""
        while True:
            # Find a chunk key to process
            chunk_key = next(self._chunk_finder, None)
            if chunk_key is None:
                self._generation_count -= 1
                return
            elif chunk_key in self._chunks or chunk_key in self._pending_chunks:
                continue
            else:
                break

        shared_chunk_data = self._shared.get_chunk(chunk_key)
        widget_chunk_data = WidgetChunkData(shared_chunk_data)
        shared_geometry = shared_chunk_data.geometry

        def create_vao(chunk: WidgetChunkData):
            if not self._context.makeCurrent(self._surface):
                raise RuntimeError("Could not make context current.")

            f = QOpenGLContext.currentContext().functions()

            if chunk.vao is None:
                # Create the VAO if one does not exist.
                chunk.vao = QOpenGLVertexArrayObject()
                chunk.vao.create()
            chunk.vao.bind()

            # Associate the vbo with the vao
            vbo_container = chunk.shared.geometry
            vbo_container.vbo.bind()

            # vertex coord
            f.glEnableVertexAttribArray(0)
            f.glVertexAttribPointer(
                0, 3, GL_FLOAT, GL_FALSE, 12 * FloatSize, VoidPtr(0)
            )
            # texture coord
            f.glEnableVertexAttribArray(1)
            f.glVertexAttribPointer(
                1, 2, GL_FLOAT, GL_FALSE, 12 * FloatSize, VoidPtr(3 * FloatSize)
            )
            # texture bounds
            f.glEnableVertexAttribArray(2)
            f.glVertexAttribPointer(
                2, 4, GL_FLOAT, GL_FALSE, 12 * FloatSize, VoidPtr(5 * FloatSize)
            )
            # tint
            f.glEnableVertexAttribArray(3)
            f.glVertexAttribPointer(
                3, 3, GL_FLOAT, GL_FALSE, 12 * FloatSize, VoidPtr(9 * FloatSize)
            )

            chunk.vao.release()
            vbo_container.vbo.release()

            # Update the vbo attribute.
            # If a VBO was previously stored it will get automatically deleted when the last reference is lost.
            chunk.geometry = vbo_container

            self.geometry_changed.emit()

        if shared_geometry is None:
            # The VBO does not exist yet. on_change will be run when it is loaded
            self._pending_chunks[chunk_key] = widget_chunk_data
            self._generation_count -= 1
        else:
            # vbo already exists
            widget_chunk_data.geometry = shared_geometry
            self._chunks[chunk_key] = widget_chunk_data
            create_vao(widget_chunk_data)
            self._generation_count -= 1
            self._queue_next_chunk()

        def on_change():
            self._generation_count += 1
            if chunk_key in self._pending_chunks:
                self._chunks[chunk_key] = self._pending_chunks.pop(chunk_key)
            if chunk_key in self._chunks:
                create_vao(self._chunks[chunk_key])

            self._generation_count -= 1
            self._queue_next_chunk()

        widget_chunk_data.geometry_changed.connect(on_change)



from __future__ import annotations

import json
import os
import re
from pathlib import Path
from typing import Any

from amulet_editor.data import build
from PySide6.QtGui import QColor
from PySide6.QtWidgets import QApplication


class Color:
    def __init__(self, colour_hex: str):
        self._qcolor = QColor(colour_hex)

    def get_qcolor(self) -> QColor:
        return self._qcolor

    def get_hex(self) -> str:
        return self._qcolor.name()

    def get_rgb(self) -> str:
        return "rgb({}, {}, {})".format(*self._qcolor.getRgb())

    def get_rgba(self) -> str:
        return "rgba({}, {}, {}, {})".format(*self._qcolor.getRgb())

    def darker(self, percent: int = 100) -> Color:
        if percent < 0:
            raise ValueError("'percent' must be a positive integer")
        return Color(self._qcolor.darker(100 + percent).name())

    def lighter(self, percent: int = 50) -> Color:
        if percent < 0:
            raise ValueError("'percent' must be a positive integer")
        return Color(self._qcolor.lighter(100 + percent).name())


class Theme:
    def __init__(self, theme_dir: str):
        with open(os.path.join(theme_dir, "theme.json"), "r") as fh:
            self._theme: dict[str, Any] = json.load(fh)

        self._style_sheets: dict[str, str] = {}

        path_style_sheets = build.get_resource(
            os.path.join("themes", "_default", "style_sheets")
        )
        for style_name in os.listdir(path_style_sheets):
            self._style_sheets[style_name] = self.load_style_sheet(
                os.path.join(path_style_sheets, style_name)
            )

        path_style_sheets = os.path.join(theme_dir, "style_sheets")
        if os.path.isdir(path_style_sheets):
            for style_name in os.listdir(path_style_sheets):
                self._style_sheets[style_name] = self.load_style_sheet(
                    os.path.join(path_style_sheets, style_name)
                )

    def apply(self, application: QApplication):
        """Apply theme to a `QtWidgets.QApplication`."""
        application.setStyle(self._theme["style"])
        application.setStyleSheet(self._style_sheets["application.qss"])

    def get_style_sheet(self, file_name: str) -> str:
        """Returns Qt style sheet from current theme with matching file name."""
        return self._style_sheets.get(file_name, "")

    def load_style_sheet(self, file_path: str) -> str:
        """Returns Qt style sheet at the provided file path. Substitutes placeholder colors with theme colors.
        \nGenerally `get_style_sheet()` should be used instead of this method.
        """
        with open(file_path, "r") as fh:
            style_sheet = fh.read().replace('"', "")

        icons: set[str] = set(
            [icon for icon in re.findall("url\((.*?)\)", style_sheet)]
        )
        for icon in icons:
            style_sheet = style_sheet.replace(
                "url({})".format(icon),
                "url({})".format(
                    Path(build.get_resource(os.path.join("icons", icon))).as_posix()
                ),
            )

        font_subfamilies: set[str] = set(
            [
                subfamily
                for subfamily in re.findall("{(.+?)}", style_sheet)
                if "font_family." in subfamily
            ]
        )
        for subfamily in font_subfamilies:
            try:
                font_data = subfamily.split(".")
                style_sheet = style_sheet.replace(
                    f"{{{subfamily}}}",
                    '"{}"'.format(
                        " ".join(
                            [
                                self.font_family,
                                self.font_subfamilies.get(font_data[1], ""),
                            ]
                        ).strip()
                    ),
                )
            except:
                raise ValueError(
                    f'unparsable font data: "{font_data}" in stylesheet "{file_path}"'
                )

        modified_colors: set[str] = set(
            [
                color
                for color in re.findall("{(.+?)}", style_sheet)
                if color not in self._theme["material_colors"]
                and color.rsplit(".", 2)[0] in self._theme["material_colors"]
            ]
        )
        for modified_color in modified_colors:
            try:
                color_data = modified_color.split(".")
                if color_data[1] == "darker":
                    style_sheet = style_sheet.replace(
                        f"{{{modified_color}}}",
                        Color(self._theme["material_colors"][color_data[0]])
                        .darker(int(color_data[2]))
                        .get_hex(),
                    )
                elif color_data[1] == "lighter":
                    style_sheet = style_sheet.replace(
                        f"{{{modified_color}}}",
                        Color(self._theme["material_colors"][color_data[0]])
                        .lighter(int(color_data[2]))
                        .get_hex(),
                    )
            except:
                raise ValueError(
                    f'unparsable color data: "{color_data}" in stylesheet "{file_path}"'
                )

        return style_sheet.format(
            font_family=self.font_family,
            background=self.background.get_hex(),
            error=self.error.get_hex(),
            primary=self.primary.get_hex(),
            primary_variant=self.primary_variant.get_hex(),
            secondary=self.secondary.get_hex(),
            secondary_variant=self.secondary_variant.get_hex(),
            surface=self.surface.get_hex(),
            on_background=self.on_background.get_hex(),
            on_error=self.on_error.get_hex(),
            on_primary=self.on_primary.get_hex(),
            on_secondary=self.on_secondary.get_hex(),
            on_surface=self.on_surface.get_hex(),
        )

    @property
    def name(self) -> str:
        return self._theme["theme_name"]

    @property
    def font_family(self) -> str:
        return self._theme["font"]["family"]

    @property
    def font_subfamilies(self) -> dict[str, str]:
        return self._theme["font"].get("subfamilies", {})

    @property
    def font_size(self) -> str:
        return self._theme["font"]["size"]

    @property
    def primary(self) -> Color:
        return Color(self._theme["material_colors"]["primary"])

    @property
    def primary_variant(self) -> Color:
        return Color(self._theme["material_colors"]["primary_variant"])

    @property
    def on_primary(self) -> Color:
        return Color(self._theme["material_colors"]["on_primary"])

    @property
    def secondary(self) -> Color:
        return Color(self._theme["material_colors"]["secondary"])

    @property
    def secondary_variant(self) -> Color:
        return Color(self._theme["material_colors"]["secondary_variant"])

    @property
    def on_secondary(self) -> Color:
        return Color(self._theme["material_colors"]["on_secondary"])

    @property
    def background(self) -> Color:
        return Color(self._theme["material_colors"]["background"])

    @property
    def on_background(self) -> Color:
        return Color(self._theme["material_colors"]["on_background"])

    @property
    def surface(self) -> Color:
        return Color(self._theme["material_colors"]["surface"])

    @property
    def on_surface(self) -> Color:
        return Color(self._theme["material_colors"]["on_surface"])

    @property
    def error(self) -> Color:
        return Color(self._theme["material_colors"]["error"])

    @property
    def on_error(self) -> Color:
        return Color(self._theme["material_colors"]["on_error"])



from typing import Tuple, Dict, Optional
import struct
import hashlib
import os
import json
import glob
import logging
from threading import Lock, RLock
from weakref import WeakKeyDictionary, ref

from PIL import Image
from PIL.ImageQt import ImageQt

from PySide6.QtCore import QObject, Signal
from PySide6.QtGui import QImage, QOpenGLContext, QOffscreenSurface
from PySide6.QtOpenGL import QOpenGLTexture

from minecraft_model_reader.api.resource_pack.base import BaseResourcePackManager
from minecraft_model_reader import BlockMesh
import PyMCTranslate
from amulet.api.block import Block
from amulet.api.level import BaseLevel

from ._textureatlas import create_atlas

from amulet_editor.application._invoke import invoke
from amulet_editor.models.widgets.traceback_dialog import DisplayException
from amulet_editor.models.generic._promise import Promise
from amulet_editor.data.paths._application import cache_directory

from amulet_team_resource_pack._api import get_resource_pack_container

log = logging.getLogger(__name__)


class OpenGLResourcePack:
    """
    This class will take a minecraft_model_reader resource pack and load the textures into a texture atlas.
    After creating an instance, initialise must be called.

    """

    _lock = Lock()
    _resource_pack: BaseResourcePackManager
    # The translator to look up the version block
    _translator: PyMCTranslate.Version
    # Loaded block models
    _block_models: Dict[Block, BlockMesh]
    # Texture coordinates
    _texture_bounds: Dict[str, Tuple[float, float, float, float]]

    # Image on GPU
    _texture: Optional[QOpenGLTexture]
    _context: Optional[QOpenGLContext]
    _surface: Optional[QOffscreenSurface]

    def __init__(
        self, resource_pack: BaseResourcePackManager, translator: PyMCTranslate.Version
    ):
        self._lock = Lock()
        self._resource_pack = resource_pack
        self._translator = translator
        self._block_models = {}
        self._texture_bounds = {}
        self._texture = None
        self._context = None
        self._surface = None

    def __del__(self):
        if self._texture is not None:
            self._context.makeCurrent(self._surface)
            self._texture.destroy()
            self._context.doneCurrent()

    def initialise(self) -> Promise[None]:
        """
        Create the atlas texture.
        """

        def func(promise_data: Promise.Data):
            with self._lock:
                if self._texture is None:
                    cache_id = struct.unpack(
                        "H",
                        hashlib.sha1(
                            "".join(self._resource_pack.pack_paths).encode("utf-8")
                        ).digest()[:2],
                    )[0]

                    atlas: Image.Image

                    if not self._resource_pack.pack_paths:
                        log.warning("There are no resource packs to load.")

                    mod_time = max(
                        (
                            os.stat(path).st_mtime
                            for pack in self._resource_pack.pack_paths
                            for path in glob.glob(
                                os.path.join(glob.escape(pack), "**", "*.*"),
                                recursive=True,
                            )
                        ),
                        default=0,
                    )

                    cache_dir = os.path.join(cache_directory(), "resource_pack")
                    img_path = os.path.join(cache_dir, f"{cache_id}.png")
                    bounds_path = os.path.join(cache_dir, f"{cache_id}.json")
                    try:
                        with open(bounds_path) as f:
                            cache_mod_time, bounds = json.load(f)
                        if mod_time != cache_mod_time:
                            raise Exception(
                                "The resource packs have changed since last merging."
                            )
                        _atlas = QImage(img_path)
                    except Exception:
                        (
                            atlas,
                            bounds,
                        ) = create_atlas(
                            self._resource_pack.textures
                        ).call_chained(promise_data)

                        os.makedirs(cache_dir, exist_ok=True)
                        atlas.save(img_path)
                        with open(bounds_path, "w") as f:
                            json.dump((mod_time, bounds), f)
                        _atlas = ImageQt(atlas)

                    self._texture_bounds = bounds

                    def init_gl():
                        self._context = QOpenGLContext()
                        self._context.setShareContext(
                            QOpenGLContext.globalShareContext()
                        )
                        self._context.create()
                        self._surface = QOffscreenSurface()
                        self._surface.create()
                        if not self._context.makeCurrent(self._surface):
                            raise RuntimeError("Could not make context current.")

                        self._texture = QOpenGLTexture(QOpenGLTexture.Target.Target2D)
                        self._texture.setMinificationFilter(
                            QOpenGLTexture.Filter.Nearest
                        )
                        self._texture.setMagnificationFilter(
                            QOpenGLTexture.Filter.Nearest
                        )
                        self._texture.setWrapMode(
                            QOpenGLTexture.CoordinateDirection.DirectionS,
                            QOpenGLTexture.WrapMode.ClampToEdge,
                        )
                        self._texture.setWrapMode(
                            QOpenGLTexture.CoordinateDirection.DirectionT,
                            QOpenGLTexture.WrapMode.ClampToEdge,
                        )
                        self._texture.setData(_atlas)
                        self._texture.create()

                        self._context.doneCurrent()

                    invoke(init_gl)

        return Promise(func)

    def get_texture(self) -> QOpenGLTexture:
        """
        Get the opengl texture for the atlas.
        The GPU data will be destroyed when the last reference to this instance is released.
        :return: A QOpenGLTexture instance.
        """
        with self._lock:
            if self._texture is None:
                raise RuntimeError("The OpenGLResourcePack has not been initialised.")
            return self._texture

    def get_texture_path(self, namespace: Optional[str], relative_path: str):
        """Get the absolute path of the image from the relative components.
        Useful for getting the id of textures for hard coded textures not connected to a resource pack.
        """
        return self._resource_pack.get_texture_path(namespace, relative_path)

    def texture_bounds(self, texture_path: str) -> Tuple[float, float, float, float]:
        """Get the bounding box of a given texture path."""
        if texture_path in self._texture_bounds:
            return self._texture_bounds[texture_path]
        else:
            return self._texture_bounds[self._resource_pack.missing_no]

    def get_block_model(self, universal_block: Block) -> BlockMesh:
        """Get the BlockMesh class for a given universal Block.
        The Block will be translated to the version format using the
        previously specified translator."""
        if universal_block not in self._block_models:
            version_block = self._translator.block.from_universal(
                universal_block.base_block
            )[0]
            if universal_block.extra_blocks:
                for block_ in universal_block.extra_blocks:
                    version_block += self._translator.block.from_universal(block_)[0]

            self._block_models[universal_block] = self._resource_pack.get_block_model(
                version_block
            )

        return self._block_models[universal_block]


class RenderResourcePackContainer(QObject):
    # Emitted with a promise object when a resource pack change is started.
    changing = Signal(object)  # Promise[None]
    # Emitted when the resource pack has changed.
    changed = Signal()

    def __init__(self, level: BaseLevel):
        super().__init__()
        self._level = ref(level)
        self._lock = RLock()
        self._resource_pack: Optional[OpenGLResourcePack] = None
        self._loader: Optional[Promise[None]] = None
        self._resource_pack_container = get_resource_pack_container(level)
        self._resource_pack_container.changed.connect(self._reload)

    @property
    def loader(self) -> Optional[Promise[None]]:
        return self._loader

    @property
    def loaded(self) -> bool:
        """Is there a valid resource pack."""
        return self._resource_pack is not None

    @property
    def resource_pack(self) -> OpenGLResourcePack:
        rp = self._resource_pack
        if rp is None:
            raise RuntimeError(
                "The OpenGLResourcePack for this level has not been loaded yet."
            )
        return rp

    def _reload(self):
        def func(promise_data: Promise.Data):
            with self._lock, DisplayException(
                "Error initialising the OpenGL resource pack."
            ):
                level: BaseLevel = self._level()
                log.debug(f"Loading OpenGL resource pack for level {level.level_path}")
                resource_pack = self._resource_pack_container.resource_pack
                # TODO: modify the resource pack library to expose the desired translator
                translator = level.translation_manager.get_version("java", (999, 0, 0))

                rp = OpenGLResourcePack(resource_pack, translator)
                promise = rp.initialise()
                # TODO: support canceling
                promise.call_chained(promise_data)
                # for progress in rp.initialise():
                #     if promise_data.is_cancel_requested or _tokens.get(level) is not token:
                #         # Abort if a new generation has been started
                #         raise Promise.OperationCanceled()
                #
                #     promise_data.progress_change.emit(0.5 + progress * 0.5)

                self._resource_pack = rp
                self.changed.emit()
                log.debug(f"Loaded OpenGL resource pack for level {level}")

        promise_ = Promise(func)
        old_loader = self._loader
        if old_loader is not None:
            old_loader.cancel()
        self._loader = promise_
        self.changing.emit(promise_)
        promise_.start()


_lock = Lock()
_level_data: WeakKeyDictionary[
    BaseLevel, RenderResourcePackContainer
] = WeakKeyDictionary()


def get_gl_resource_pack_container(level: BaseLevel) -> RenderResourcePackContainer:
    with _lock:
        if level not in _level_data:
            _level_data[level] = invoke(lambda: RenderResourcePackContainer(level))
        return _level_data[level]



# -*- coding: utf-8 -*-
################################################################################
## Form generated from reading UI file '_settings.ui'
##
## Created by: Qt User Interface Compiler version 6.5.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################
from PySide6.QtCore import QCoreApplication, QMetaObject, QEvent
from PySide6.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)


class Ui_SettingsPage(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.objectName():
            self.setObjectName("SettingsPage")
        self.resize(748, 788)

        self.horizontalLayout = QHBoxLayout(self)
        self.horizontalLayout.setSpacing(5)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)

        self.horizontalSpacer_2 = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )
        self.horizontalLayout.addItem(self.horizontalSpacer_2)

        self._central_layout = QVBoxLayout()
        self._central_layout.setObjectName("_central_layout")
        self._central_layout.setContentsMargins(50, 50, 50, 50)

        self.label = QLabel(self)
        self.label.setObjectName("label")
        self._central_layout.addWidget(self.label)

        self.verticalSpacer = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding
        )
        self._central_layout.addItem(self.verticalSpacer)
        self.horizontalLayout.addLayout(self._central_layout)

        self.horizontalSpacer = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )
        self.horizontalLayout.addItem(self.horizontalSpacer)
        self.horizontalLayout.setStretch(0, 1)
        self.horizontalLayout.setStretch(2, 1)

        self.localise()
        QMetaObject.connectSlotsByName(self)

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() == QEvent.Type.LanguageChange:
            self.localise()

    def localise(self):
        self.setWindowTitle(QCoreApplication.translate("SettingsPage", "Form", None))
        self.label.setText(QCoreApplication.translate("SettingsPage", "Settings", None))



import re

from amulet_editor.application import appearance
from PySide6.QtCore import QRect, QSize, Qt
from PySide6.QtGui import (
    QColor,
    QKeyEvent,
    QPainter,
    QPaintEvent,
    QResizeEvent,
    QSyntaxHighlighter,
    QTextCharFormat,
    QTextFormat,
)
from PySide6.QtWidgets import QPlainTextEdit, QTextEdit, QWidget


class QCodeEditor(QPlainTextEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.lineNumberArea = QLineNumberArea(self)
        self.lineNumberArea.setProperty("backgroundColor", "background")
        self.lineNumberArea.setProperty("borderRight", "surface")
        self.lineNumberArea.setProperty("color", "on_background")

        self.blockCountChanged.connect(self.updateLineNumberAreaWidth)
        self.updateRequest.connect(self.updateLineNumberArea)
        self.cursorPositionChanged.connect(self.highlightCurrentLine)
        self.updateLineNumberAreaWidth(0)

    def lineNumberAreaWidth(self):
        digits = 1
        max_value = max(1, self.blockCount())
        while max_value >= 10:
            max_value /= 10
            digits += 1
        space = 3 + self.fontMetrics().horizontalAdvance("9") * (digits + 2)
        return space

    def updateLineNumberAreaWidth(self, _):
        self.setViewportMargins(self.lineNumberAreaWidth(), 0, 0, 0)

    def updateLineNumberArea(self, rect: QRect, dy: int):
        if dy:
            self.lineNumberArea.scroll(0, dy)
        else:
            self.lineNumberArea.update(
                0, rect.y(), self.lineNumberArea.width(), rect.height()
            )
        if rect.contains(self.viewport().rect()):
            self.updateLineNumberAreaWidth(0)

    def resizeEvent(self, event: QResizeEvent):
        super().resizeEvent(event)
        cr = self.contentsRect()
        self.lineNumberArea.setGeometry(
            QRect(cr.left(), cr.top(), self.lineNumberAreaWidth(), cr.height())
        )

    def highlightCurrentLine(self):
        extraSelections = []
        if not self.isReadOnly():
            selection = QTextEdit.ExtraSelection()
            lineColor = appearance.theme().primary_variant.get_qcolor()
            selection.format.setBackground(lineColor)
            selection.format.setProperty(QTextFormat.Property.FullWidthSelection, True)
            selection.cursor = self.textCursor()
            selection.cursor.clearSelection()
            extraSelections.append(selection)
        self.setExtraSelections(extraSelections)

    def lineNumberAreaPaintEvent(self, event: QPaintEvent):
        painter = QPainter(self.lineNumberArea)

        painter.fillRect(event.rect(), appearance.theme().background.get_qcolor())

        block = self.firstVisibleBlock()
        blockNumber = block.blockNumber()
        top = self.blockBoundingGeometry(block).translated(self.contentOffset()).top()
        bottom = top + self.blockBoundingRect(block).height()

        height = self.fontMetrics().height()
        while block.isValid() and (top <= event.rect().bottom()):
            if block.isVisible() and (bottom >= event.rect().top()):
                number = str(blockNumber + 1) + " "
                painter.setPen(appearance.theme().on_surface.get_qcolor())
                painter.drawText(
                    0,
                    top,
                    self.lineNumberArea.width(),
                    height,
                    Qt.AlignmentFlag.AlignRight,
                    number,
                )

            block = block.next()
            top = bottom
            bottom = top + self.blockBoundingRect(block).height()
            blockNumber += 1

    def keyPressEvent(self, event: QKeyEvent):
        if event.key() == Qt.Key.Key_Tab:
            tc = self.textCursor()
            tc.insertText("    ")
        else:
            super().keyPressEvent(event)


class QLineNumberArea(QWidget):
    def __init__(self, editor: QCodeEditor):
        super().__init__(editor)
        self.editor = editor

    def sizeHint(self):
        return QSize(self.editor.lineNumberAreaWidth(), 0)

    def paintEvent(self, event: QPaintEvent):
        self.editor.lineNumberAreaPaintEvent(event)


class MCFunctionHighlighter(QSyntaxHighlighter):
    def __init__(self, parent=None):
        super().__init__(parent)

        self._mappings = {}

        comment_format = QTextCharFormat()
        comment_format.setForeground(QColor("#529955"))
        pattern = r"^\s*#.*$"
        self.add_mapping(pattern, comment_format)

        function_list = [
            "advancement",
            "attribute",
            "ban",
            "ban-ip",
            "banlist",
            "bossbar",
            "clear",
            "clone",
            "data",
            "datapack",
            "debug",
            "defaultgamemode",
            "deop",
            "difficulty",
            "effect",
            "enchant",
            "execute",
            "experience",
            "fill",
            "forceload",
            "function",
            "gamemode",
            "gamerule",
            "give",
            "help",
            "item",
            "kick",
            "kill",
            "list",
            "locate",
            "locatebiome",
            "loot",
            "me",
            "msg",
            "op",
            "pardon",
            "pardon-ip",
            "particle",
            "perf",
            "place",
            "playsound",
            "publish",
            "recipe",
            "reload",
            "save-all",
            "save-off",
            "save-on",
            "say",
            "schedule",
            "scoreboard",
            "seed",
            "setblock",
            "setidletimeout",
            "setworldspawn",
            "spawnpoint",
            "spectate",
            "spreadplayers",
            "stop",
            "stopsound",
            "summon",
            "tag",
            "team",
            "teammsg",
            "teleport",
            "tell",
            "tellraw",
            "time",
            "title",
            "tm",
            "tp",
            "trigger",
            "warden_spawn_tracker",
            "weather",
            "whitelist",
            "worldborder",
            "xp",
        ]
        function_re = "|".join(function_list)
        function_format = QTextCharFormat()
        function_format.setForeground(QColor("#c586c0"))
        pattern = r"^(" + function_re + r")\b|(?<=run )\b(" + function_re + r")\b"
        self.add_mapping(pattern, function_format)

    def add_mapping(self, pattern, fmt):
        self._mappings[pattern] = fmt

    def highlightBlock(self, text):
        for pattern, fmt in self._mappings.items():
            for match in re.finditer(pattern, text):
                start, end = match.span()
                self.setFormat(start, end - start, fmt)


class JsonHighlighter(QSyntaxHighlighter):
    def __init__(self, parent=None):
        super().__init__(parent)

        self._mappings = {}

        bool_format = QTextCharFormat()
        bool_format.setForeground(QColor("#569cd6"))
        pattern = r"([tT]rue|[fF]alse)"
        self.add_mapping(pattern, bool_format)

        string_format = QTextCharFormat()
        string_format.setForeground(QColor("#f4ab6e"))
        pattern = r"\"(.*?)(?<!\\)\""
        self.add_mapping(pattern, string_format)

        key_format = QTextCharFormat()
        key_format.setForeground(QColor("#9cdcfe"))
        pattern = r"(\".*\")(?=\s*:)"
        self.add_mapping(pattern, key_format)

        string_value_format = QTextCharFormat()
        string_value_format.setForeground(QColor("#f4ab6e"))
        pattern = r"(?<=:\s)([\"\'].*[\"\'])"
        self.add_mapping(pattern, string_value_format)

        number_value_format = QTextCharFormat()
        number_value_format.setForeground(QColor("#a7cea8"))
        pattern = r"(?<=:\s)(-*((\d+[.]\d+)|(\d+)))"
        self.add_mapping(pattern, number_value_format)

    def add_mapping(self, pattern, fmt):
        self._mappings[pattern] = fmt

    def highlightBlock(self, text):
        for pattern, fmt in self._mappings.items():
            for match in re.finditer(pattern, text):
                start, end = match.span()
                self.setFormat(start, end - start, fmt)



from __future__ import annotations

from enum import IntEnum


class PluginState(IntEnum):
    Disabled = 0  # Plugin is not enabled by the user
    Inactive = (
        1  # Plugin is enabled by the user but had dependencies that are not enabled
    )
    Enabled = 2  # Plugin is fully enabled



import os
from typing import Optional

from amulet_editor.application.appearance._theme import Theme
from amulet_editor.data import build, project
from PySide6.QtCore import QObject, Signal


class ThemeManager(QObject):
    changed = Signal(object)

    def __init__(self):
        super().__init__(parent=None)
        self._theme: Optional[Theme] = None

        self._themes: list[Theme] = []
        theme_dir = build.get_resource("themes")
        for theme_ in os.listdir(theme_dir):
            if theme_ != "_default":
                self._themes.append(Theme(os.path.join(theme_dir, theme_)))

        self.set_theme(project.settings()["theme"])

    @property
    def theme(self) -> Theme:
        return self._theme

    def list_themes(self) -> list[str]:
        return [theme_.name for theme_ in self._themes]

    def set_theme(self, theme_name: str):
        for theme_ in self._themes:
            if theme_.name == theme_name:
                self._theme = theme_
                self.changed.emit(self.theme)
                return


theme_manager = ThemeManager()

changed = theme_manager.changed


def list_themes() -> list[str]:
    return theme_manager.list_themes()


def set_theme(theme_name: str):
    return theme_manager.set_theme(theme_name)


def theme() -> Theme:
    return theme_manager.theme



from typing import Optional

from PySide6.QtCore import QCoreApplication, Qt, Signal
from PySide6.QtWidgets import (
    QFrame,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QScrollArea,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)


class QMenuWidget(QWidget):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.setupUi()

    @property
    def clicked_cancel(self) -> Signal:
        return self.btn_cancel.clicked

    @property
    def clicked_back(self) -> Signal:
        return self.btn_back.clicked

    @property
    def clicked_next(self) -> Signal:
        return self.btn_next.clicked

    def setMenuTitle(self, text: str):
        self._lbl_menu_title.setText(text)

    def menuTitle(self) -> str:
        return self._lbl_menu_title.text()

    def setWidget(self, widget: QWidget):
        if widget not in self._swgt_menu.children():
            self._swgt_menu.addWidget(widget)
        self._swgt_menu.setCurrentWidget(widget)

    def removeWidget(self, widget: QWidget):
        self._swgt_menu.removeWidget(widget)

    def setupUi(self):
        # Create 'Menu' frame
        frm_menu = QFrame(self)
        frm_menu.setFrameShape(QFrame.Shape.NoFrame)
        frm_menu.setFrameShadow(QFrame.Shadow.Raised)
        frm_menu.setMaximumWidth(750)

        # Create 'Header' frame
        frm_menu_header = QFrame(frm_menu)

        lbl_menu_header = QLabel(frm_menu_header)
        lbl_menu_header.setProperty("heading", "h3")
        lbl_menu_header.setProperty("subfamily", "semi_light")

        lyt_menu_header = QHBoxLayout(frm_menu_header)
        lyt_menu_header.addWidget(lbl_menu_header)
        lyt_menu_header.setAlignment(Qt.AlignmentFlag.AlignLeft)
        lyt_menu_header.setContentsMargins(0, 0, 0, 5)
        lyt_menu_header.setSpacing(5)

        frm_menu_header.setFrameShape(QFrame.Shape.NoFrame)
        frm_menu_header.setFrameShadow(QFrame.Shadow.Raised)
        frm_menu_header.setLayout(lyt_menu_header)

        # Create swappable 'Menu Container' widget
        swgt_menu = QStackedWidget()
        swgt_menu.setProperty("backgroundColor", "background")

        # Create 'Menu Container' widget
        scr_menu = QScrollArea(frm_menu)
        scr_menu.setFrameShape(QFrame.Shape.NoFrame)
        scr_menu.setFrameShadow(QFrame.Shadow.Raised)
        scr_menu.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        scr_menu.setProperty("borderBottom", "surface")
        scr_menu.setProperty("borderTop", "surface")
        scr_menu.setProperty("backgroundColor", "background")
        scr_menu.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        scr_menu.setWidgetResizable(True)
        scr_menu.setWidget(swgt_menu)

        # Create 'Page Options' frame
        frm_menu_options = QFrame(frm_menu)

        # Configure 'Menu Options' widgets
        btn_cancel = QPushButton(frm_menu_options)
        btn_cancel.setFixedHeight(30)

        btn_back = QPushButton(frm_menu_options)
        btn_back.setFixedHeight(30)

        btn_next = QPushButton(frm_menu_options)
        btn_next.setFixedHeight(30)
        btn_next.setProperty("backgroundColor", "secondary")

        # Create 'Menu Options' layout
        lyt_menu_options = QHBoxLayout(frm_menu)
        lyt_menu_options.addWidget(btn_cancel)
        lyt_menu_options.addStretch()
        lyt_menu_options.addWidget(btn_back)
        lyt_menu_options.addWidget(btn_next)
        lyt_menu_options.setContentsMargins(0, 10, 0, 5)
        lyt_menu_options.setSpacing(5)

        # Configure 'Menu Options' frame
        frm_menu_options.setFrameShape(QFrame.Shape.NoFrame)
        frm_menu_options.setFrameShadow(QFrame.Shadow.Raised)
        frm_menu_options.setLayout(lyt_menu_options)

        # Create 'Menu' layout
        lyt_menu = QVBoxLayout(frm_menu)
        lyt_menu.addWidget(frm_menu_header)
        lyt_menu.addWidget(scr_menu)
        lyt_menu.addWidget(frm_menu_options)
        lyt_menu.setAlignment(Qt.AlignmentFlag.AlignCenter)
        lyt_menu.setSpacing(5)

        layout = QVBoxLayout(self)
        layout.addWidget(frm_menu)
        layout.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        self.setLayout(layout)

        self._lbl_menu_title = lbl_menu_header
        self._swgt_menu = swgt_menu
        self.btn_cancel = btn_cancel
        self.btn_back = btn_back
        self.btn_next = btn_next

        self.retranslateUi()

    def retranslateUi(self):
        # Disable formatting to condense tranlate functions
        # fmt: off
        self.btn_cancel.setText(QCoreApplication.translate("QMenuWidget", "Cancel", None))
        self.btn_back.setText(QCoreApplication.translate("QMenuWidget", "Back", None))
        self.btn_next.setText(QCoreApplication.translate("QMenuWidget", "Next", None))
        # fmt: on



from __future__ import annotations

import re
from packaging.specifiers import SpecifierSet

from ._uid import LibraryUID


RequirementPattern = re.compile(
    r"(?P<identifier>[a-zA-Z_]+[a-zA-Z_0-9]*)(?P<requirement>.*)"
)


class Requirement:
    def __init__(self, identifier: str, specifier: SpecifierSet):
        self._identifier = identifier.lower().replace("-", "_")
        self._specifier = specifier

    @property
    def identifier(self) -> str:
        """The package name"""
        return self._identifier

    @property
    def specifier(self) -> SpecifierSet:
        """The version specifier. It is recommended to use the compatible format. Eg. "~=1.0"""
        return self._specifier

    @classmethod
    def from_string(cls, requirement: str):
        match = RequirementPattern.fullmatch(requirement)
        if match is None:
            raise ValueError(
                f'"{requirement}" is not a valid requirement.\n It must be a python identifier followed by an optional PEP 440 compatible version specifier'
            )
        specifier = SpecifierSet(match.group("requirement"))
        return cls(match.group("identifier"), specifier)

    def __contains__(self, item: LibraryUID):
        if not isinstance(item, LibraryUID):
            raise TypeError
        return item.identifier == self.identifier and item.version in self.specifier

    def __str__(self):
        return f"{self.identifier}{self.specifier}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.identifier!r}, {self.specifier!r})"



import json
import os
from os import R_OK
from pathlib import Path
from typing import Optional

from amulet_editor.tools.project._components import (
    JsonHighlighter,
    MCFunctionHighlighter,
    QCodeEditor,
)
from PySide6.QtGui import QFontDatabase, QSyntaxHighlighter
from PySide6.QtWidgets import QVBoxLayout, QWidget

syntax_highlighters = {
    ".mcfunction": MCFunctionHighlighter,
    ".json": JsonHighlighter,
    ".mcmeta": JsonHighlighter,
}

json_like = (".json", ".mcmeta")


class ProjectPage(QWidget):
    def __init__(self, parent: Optional[QWidget] = None):
        if parent is None:
            super().__init__()
        else:
            super().__init__(parent=parent)

        self.highlighter: Optional[QSyntaxHighlighter] = None

        self.setupUi()

    def show_file(self, file_path: str):
        if os.path.isfile(file_path) and os.access(file_path, R_OK):
            try:
                with open(file_path, "r") as textio:
                    if Path(file_path).suffix in json_like:
                        parsed = json.load(textio)
                        text = json.dumps(parsed, indent=2)
                    else:
                        text = textio.read()

                self.txe_file.setPlainText(text)

                # Remove current syntax highlighter
                if self.highlighter is not None:
                    self.highlighter.setDocument(None)

                # Attach new syntax highlighter to document
                highlighter = syntax_highlighters.get(Path(file_path).suffix, None)
                if highlighter is not None:
                    self.highlighter = highlighter()
                    self.highlighter.setDocument(self.txe_file.document())

            except UnicodeDecodeError:
                self.txe_file.clear()
                self.txe_file.setPlainText(f"Error Decoding File:\n{file_path}\n")

    def setupUi(self):
        font = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)

        self.txe_file = QCodeEditor()
        self.txe_file.setFont(font)

        # self.hlt_mcfunction = MCFunctionHighlighter()
        # self.hlt_mcfunction.setDocument(self.txe_file.document())
        # self.hlt_mcfunction.setDocument(None)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.txe_file)

        self.setLayout(layout)



from __future__ import annotations
from typing import Type, Callable, TYPE_CHECKING
from uuid import uuid4

from PySide6.QtGui import QShortcut
from PySide6.QtCore import Qt

from ._landing_window import Ui_AmuletLandingWindow
from ._view import ViewContainer, View

from amulet_team_inspector import show_inspector

if TYPE_CHECKING:
    from amulet_editor.models.widgets._toolbar import AToolBar


# Terminology
# A tool button is a button in the toolbar
# A view is a program that exists within Amulet
# A view container is an object that can contain a view


UID = UUID = str


class AmuletMainWindow(Ui_AmuletLandingWindow):
    _toolbar: AToolBar

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._view_classes: dict[Type[View], UUID] = {}
        self._view_containers: list[ViewContainer] = [self._view_container]
        self._active_view = self._view_container
        f12 = QShortcut(Qt.Key.Key_F12, self)
        f12.activated.connect(show_inspector)

    def activate_view(self, view_cls: Type[View]):
        """
        Enable a view registered to the given identifier.
        If a view of this type already exists it will be activated otherwise a new view will be created in the active container.

        :param view_cls: The view type to enable
        """
        if view_cls not in self._view_classes:
            raise ValueError(f"View type {view_cls} has not been registered")

        # Scan the view containers to see if a view of this type already exists
        for existing_view in self._view_containers:
            if existing_view.view_cls is view_cls:
                break
        else:
            existing_view = None

        if existing_view is None:
            # Replace the active view
            view = view_cls()
            old_view = self._view_container.swap_view(view)
            # TODO: implement view caching
            if old_view is not None:
                old_view.deleteLater()
        else:
            existing_view.get_view().activate_view()

        self._toolbar.activate(self._view_classes[view_cls])

    def register_view(self, view_cls: Type[View], icon: str, name: str):
        """
        Register a view class for this window only. This adds a button to activate the view.

        :param view_cls: The view class to register.
        :param icon: The icon path to use in the toolbar.
        :param name: The name of the view to use in the icon tooltip.
        """
        if not issubclass(view_cls, View) or view_cls is View:
            raise TypeError("view must be a subclass of View")
        if view_cls in self._view_classes:
            raise ValueError(f"View type {view_cls} has already been registered.")
        self._view_classes[view_cls] = uid = str(uuid4())
        self._toolbar.add_dynamic_button(
            uid, icon, name, lambda: self.activate_view(view_cls)
        )

    def unregister_view(self, view_cls: Type[View]):
        """
        Unregister a view for this window only.

        :param view_cls: The view class to unregister.
        :return:
        """
        # TODO: raise NotImplementedError
        pass

    def add_toolbar_button(
        self, uid: UID, icon: str, name: str, callback: Callable[[], None] = None
    ):
        """
        Add an icon to the toolbar for this window only.

        :param uid: The unique identifier of the button.
        :param icon: The icon path to use in the toolbar.
        :param name: The name of the view to use in the icon tooltip.
        :param callback: The function to call when the button is clicked.
        :return:
        """
        self._toolbar.add_dynamic_button(uid, icon, name, callback)

    def add_static_toolbar_button(
        self, uid: UID, icon: str, name: str, callback: Callable[[], None] = None
    ):
        """
        Add a static icon to the toolbar for this window only.
        These should be reserved for special cases.
        Third party plugins should use :meth:`add_button`.

        :param uid: The unique identifier of the button.
        :param icon: The icon path to use in the toolbar.
        :param name: The name of the view to use in the icon tooltip.
        :param callback: The function to call when the button is clicked.
        :return:
        """
        self._toolbar.add_static_button(uid, icon, name, callback)

    def remove_toolbar_button(self, uid: UID):
        """
        Remove a toolbar button.

        :param uid: The unique identifier for the toolbar button to remove.
        :return:
        """
        self._toolbar.remove_button(uid)



from PySide6.QtCore import Qt
from PySide6.QtGui import QMouseEvent
from PySide6.QtWidgets import QHBoxLayout, QVBoxLayout, QWidget


class ADragContainer(QWidget):
    """
    Generic list sorting handler.
    """

    # orderChanged = Signal(list)

    def __init__(
        self,
        parent: QWidget = None,
        orientation: Qt.Orientation = Qt.Orientation.Vertical,
    ):
        super().__init__(parent)
        self.setMouseTracking(True)

        self._layout = {
            Qt.Orientation.Horizontal: QHBoxLayout,
            Qt.Orientation.Vertical: QVBoxLayout,
        }[orientation](self)
        self._layout.setContentsMargins(0, 0, 0, 0)

        self._drag = None

    def mouseMoveEvent(self, event: QMouseEvent):
        if event.buttons() == Qt.MouseButton.LeftButton:
            if self._drag is None:
                for i in range(self._layout.count()):
                    child = self._layout.itemAt(i).widget()
                    if child.underMouse():
                        self._drag = child
                        break
            if self._drag is not None:
                point = event.position().toPoint()
                for i in range(self._layout.count()):
                    child = self._layout.itemAt(i).widget()
                    rect = child.rect()
                    rect.moveTo(child.pos())
                    if rect.contains(point):
                        self._layout.removeWidget(self._drag)
                        self._layout.insertWidget(i, self._drag)
                        break
        else:
            self._drag = None

    def add_item(self, item: QWidget):
        self._layout.addWidget(item)



from typing import Optional

from amulet_editor.data import build
from PySide6.QtCore import QEvent, Qt
from PySide6.QtGui import QEnterEvent, QPixmap
from PySide6.QtWidgets import (
    QFrame,
    QHBoxLayout,
    QLabel,
    QLayout,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from ._icon import AStylableSvgWidget
from ._label import QElidedLabel


class QPixCard(QPushButton):
    def __init__(self, pixmap: QPixmap, parent: Optional[QWidget] = None):
        super().__init__(parent=parent)

        self._pixmap = QLabel()
        self._pixmap.setPixmap(pixmap)
        self._pixmap.setProperty("backgroundColor", "background")
        self._pixmap.setProperty("border", "surface")
        self._pixmap.setProperty("borderRadiusVisible", True)
        self._pixmap.setSizePolicy(
            QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum
        )

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        frame = QFrame()
        frame.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)
        frame.setLayout(layout)

        layout = QHBoxLayout()
        layout.setSizeConstraint(QLayout.SizeConstraint.SetMinimumSize)
        layout.addWidget(self._pixmap)
        layout.addWidget(frame)
        self.setLayout(layout)

    def addLabel(self, text: str):
        label = QElidedLabel(text, self)
        self.layout().itemAt(1).widget().layout().addWidget(label)

    def layout(self) -> QHBoxLayout:
        return super().layout()


class ALinkCard(QPushButton):
    svg_link_icon: Optional[AStylableSvgWidget]

    def __init__(
        self,
        text: str,
        icon: Optional[str] = None,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent=parent)
        self.setProperty("hover", "false")  # X:hover > Y style is broken.

        layout = QHBoxLayout()

        if icon:
            self.svg_link_icon = AStylableSvgWidget(icon)
            self.svg_link_icon.setAttribute(
                Qt.WidgetAttribute.WA_TransparentForMouseEvents
            )
            self.svg_link_icon.setFixedSize(15, 15)
            layout.addWidget(self.svg_link_icon)
        else:
            self.svg_link_icon = None

        self.lbl_description = QElidedLabel(text)
        self.lbl_description.setAttribute(
            Qt.WidgetAttribute.WA_TransparentForMouseEvents
        )
        self.lbl_description.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Maximum
        )
        layout.addWidget(self.lbl_description)

        self.svg_ext_icon = AStylableSvgWidget(
            build.get_resource(f"icons/tabler/external-link.svg")
        )
        self.svg_ext_icon.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)
        self.svg_ext_icon.setFixedSize(15, 15)
        self.svg_ext_icon.setHidden(True)
        layout.addWidget(self.svg_ext_icon)

        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.setContentsMargins(5, 5, 5, 5)
        layout.setSpacing(5)

        self.setLayout(layout)
        self.setMinimumSize(self.minimumSizeHint())

    def enterEvent(self, event: QEnterEvent):
        self.svg_ext_icon.setHidden(False)
        self.setProperty("hover", "true")
        self.setStyleSheet("/* /")  # Force a style update.
        super().enterEvent(event)

    def leaveEvent(self, event: QEvent):
        self.svg_ext_icon.setHidden(True)
        self.setProperty("hover", "false")
        self.setStyleSheet("/* /")  # Force a style update.
        super().leaveEvent(event)



# -*- coding: utf-8 -*-
################################################################################
## Form generated from reading UI file '_traceback_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.5.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################
from PySide6.QtCore import QCoreApplication, QMetaObject, QSize, Qt, QEvent
from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QTextEdit,
    QVBoxLayout,
)


class Ui_AmuletTracebackDialog(QDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.objectName():
            self.setObjectName("AmuletTracebackDialog")
        self.resize(400, 300)

        self._main_layout = QVBoxLayout(self)
        self._main_layout.setObjectName("_main_layout")

        self._header_layout = QHBoxLayout()
        self._header_layout.setObjectName("_header_layout")

        self._alert_image = QLabel(self)
        self._alert_image.setObjectName("_alert_image")
        self._alert_image.setMinimumSize(QSize(32, 32))
        self._alert_image.setMaximumSize(QSize(32, 32))
        self._header_layout.addWidget(self._alert_image)

        self._error_text = QLabel(self)
        self._error_text.setObjectName("_error_text")
        self._header_layout.addWidget(self._error_text)
        self._header_layout.setStretch(1, 1)
        self._main_layout.addLayout(self._header_layout)

        self._traceback_text = QTextEdit(self)
        self._traceback_text.setObjectName("_traceback_text")
        self._traceback_text.setReadOnly(True)
        self._main_layout.addWidget(self._traceback_text)

        self._button_layout = QHBoxLayout()
        self._button_layout.setObjectName("_button_layout")

        self._spacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self._button_layout.addItem(self._spacer)

        self._copy_button = QPushButton(self)
        self._copy_button.setObjectName("_copy_button")
        self._button_layout.addWidget(self._copy_button)

        self._ok_button_box = QDialogButtonBox(self)
        self._ok_button_box.setObjectName("_ok_button_box")
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self._ok_button_box.sizePolicy().hasHeightForWidth()
        )
        self._ok_button_box.setSizePolicy(sizePolicy)
        self._ok_button_box.setOrientation(Qt.Horizontal)
        self._ok_button_box.setStandardButtons(QDialogButtonBox.Ok)
        self._button_layout.addWidget(self._ok_button_box)
        self._main_layout.addLayout(self._button_layout)

        self.localise()
        self._ok_button_box.rejected.connect(self.reject)
        self._ok_button_box.accepted.connect(self.accept)
        QMetaObject.connectSlotsByName(self)

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() == QEvent.Type.LanguageChange:
            self.localise()

    def localise(self):
        self.setWindowTitle(
            QCoreApplication.translate("AmuletTracebackDialog", "window_title", None)
        )
        self._alert_image.setText("")
        self._error_text.setText("")
        self._copy_button.setText(
            QCoreApplication.translate("AmuletTracebackDialog", "copy_error", None)
        )



from typing import Optional

from PySide6.QtCore import QObject, Signal
from PySide6.QtWidgets import QWidget


class AmuletView(QObject):
    changed = Signal(QWidget)

    def __init__(self, widget: Optional[QWidget] = None):
        super().__init__()

        self.__widget__ = widget

    def widget(self) -> Optional[QWidget]:
        return self.__widget__

    def setWidget(self, widget: QWidget):
        self.__widget__ = widget
        self.changed.emit(self.__widget__)



from __future__ import annotations

import re

from packaging.version import Version
from runtime_final import final


@final
class LibraryUID:
    """A named tuple containing an identifier for a library/plugin and a version number."""

    def __init__(self, identifier: str, version: Version):
        self._identifier = identifier.lower().replace("-", "_")
        self._version = version

    def __repr__(self) -> str:
        return f"LibraryUID({self._identifier!r}, {self._version!r})"

    def __str__(self):
        return f"{self._identifier} {self._version}"

    @property
    def identifier(self) -> str:
        """The package name. This is the name used when importing the package. Eg "my_name_my_plugin_v1". This must be a valid python identifier."""
        return self._identifier

    @property
    def version(self) -> Version:
        """The version number of the plugin."""
        return self._version

    def to_string(self):
        return f"{self.identifier}@{self.version}"

    @classmethod
    def from_string(cls, s: str):
        match = re.fullmatch(r"(?P<identifier>[a-zA-Z_]+\w*)@(?P<version>.*)", s)
        if match is None:
            raise ValueError(f"Invalid LibraryUID string: {s}")
        version = Version(match.group("version"))
        return cls(match.group("identifier"), version)



from __future__ import annotations
import os
from typing import Optional

from PySide6.QtCore import QLocale, QCoreApplication

from amulet_editor.models.localisation import ATranslator

import amulet_team_locale


from amulet_team_main_window.api import (
    register_view,
    unregister_view,
    get_active_window,
)

import amulet_team_home_page
from .home import HomeView


# Qt only weekly references this. We must hold a strong reference to stop it getting garbage collected
_translator: Optional[ATranslator] = None


def load_plugin():
    global _translator
    _translator = ATranslator()
    _locale_changed()
    QCoreApplication.installTranslator(_translator)
    amulet_team_locale.locale_changed.connect(_locale_changed)

    register_view(HomeView, "home.svg", "Home")
    get_active_window().activate_view(HomeView)


def _locale_changed():
    _translator.load_lang(
        QLocale(),
        "",
        directory=os.path.join(*amulet_team_home_page.__path__, "resources", "lang"),
    )


def unload_plugin():
    unregister_view(HomeView)
    QCoreApplication.removeTranslator(_translator)



from __future__ import annotations

from typing import NamedTuple, TYPE_CHECKING
from packaging.specifiers import SpecifierSet

if TYPE_CHECKING:
    from ._uid import LibraryUID
    from ._requirement import Requirement


class PluginDataDepends(NamedTuple):
    """Static, publicly accessible data about a plugin's dependencies"""

    python: SpecifierSet
    library: tuple[Requirement, ...]
    plugin: tuple[
        Requirement, ...
    ]  # The plugins that this plugin depends on. This plugin will only be loaded once these plugins have been loaded.


class PluginData(NamedTuple):
    """Static, publicly accessible data about a plugin"""

    uid: LibraryUID  # The unique identifier for the plugin. Made up of the plugin id and the version number. No two plugins may have the same UID.
    path: str  # The root path of the plugin.
    name: str  # The public name of the plugin.
    depends: PluginDataDepends  # The dependencies for the plugin
    locked: bool  # Is the plugin locked on. Only accessible to first party plugins.



# -*- coding: utf-8 -*-
################################################################################
## Form generated from reading UI file '_landing_window.ui'
##
## Created by: Qt User Interface Compiler version 6.5.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################
from PySide6.QtCore import QCoreApplication, QMetaObject, QEvent
from PySide6.QtWidgets import QFrame, QHBoxLayout, QMainWindow, QWidget
from amulet_editor.application.windows._amulet_landing_window._view import ViewContainer
from amulet_editor.models.widgets._toolbar import AToolBar


class Ui_AmuletLandingWindow(QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.objectName():
            self.setObjectName("AmuletLandingWindow")
        self.resize(800, 600)

        self.central_widget = QWidget(self)
        self.central_widget.setObjectName("central_widget")

        self.horizontalLayout = QHBoxLayout(self.central_widget)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)

        self._toolbar = AToolBar(self.central_widget)
        self._toolbar.setObjectName("_toolbar")
        self._toolbar.setFrameShape(QFrame.NoFrame)
        self._toolbar.setFrameShadow(QFrame.Raised)
        self._toolbar.setProperty("backgroundColor", "surface")
        self.horizontalLayout.addWidget(self._toolbar)

        self._view_container = ViewContainer(self.central_widget)
        self._view_container.setObjectName("_view_container")
        self.horizontalLayout.addWidget(self._view_container)
        self.horizontalLayout.setStretch(1, 1)
        self.setCentralWidget(self.central_widget)

        self.localise()
        QMetaObject.connectSlotsByName(self)

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() == QEvent.Type.LanguageChange:
            self.localise()

    def localise(self):
        self.setWindowTitle(
            QCoreApplication.translate("AmuletLandingWindow", "MainWindow", None)
        )



from __future__ import annotations

from PySide6.QtCore import QCoreApplication, QMetaObject, QEvent
from PySide6.QtWidgets import QMainWindow

import amulet_team_main_window2.application.widgets.tab_engine._tab_engine as tab_engine


class AmuletSubWindow(QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.objectName():
            self.setObjectName("AmuletSubWindow")
        self.resize(800, 600)

        self.splitter_widget = tab_engine.RecursiveSplitter()
        tw = tab_engine.TabEngineStackedTabWidget()
        self.splitter_widget.addWidget(tw)
        self.setCentralWidget(self.splitter_widget)

        self.localise()
        QMetaObject.connectSlotsByName(self)

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() == QEvent.Type.LanguageChange:
            self.localise()

    def localise(self):
        self.setWindowTitle(
            QCoreApplication.translate("AmuletSubWindow", "SubWindow", None)
        )



from __future__ import annotations

from typing import Optional
import traceback

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QVBoxLayout, QLabel

from amulet_team_main_window.application.windows.main_window import View

from ._renderer import FirstPersonCanvas


class View3D(QWidget, View):
    def __init__(
        self, parent: Optional[QWidget] = None, f: Qt.WindowType = Qt.WindowType.Widget
    ):
        super().__init__(parent, f)
        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(FirstPersonCanvas())



import json as jsonlib
import pathlib
from typing import Any


class AmuletProject:
    def __init__(self, path: str):
        ext = pathlib.Path(path).suffix
        if ext == ".amulet-project":
            with open(path) as _amulet_project:
                self.__json: dict[str, dict] = jsonlib.load(_amulet_project)
            self.__path = path
        else:
            raise ValueError(
                f"Expected file extension '.amulet-project', found '{ext}'"
            )

    @property
    def packages(self) -> dict[str, list]:
        return self.__json.get("packages", {})

    @property
    def settings(self) -> dict[str, Any]:
        return self.__json.get("settings", {})

    def read(self) -> str:
        return jsonlib.dumps(self.__json, indent=4)

    def write(self, json: str):
        self.__json = jsonlib.loads(json)
        with open(self.__path) as _amulet_project:
            jsonlib.dump(self.__path, _amulet_project, indent=4)



from typing import NamedTuple, Optional
from PySide6.QtCore import QObject, Signal
from PySide6.QtGui import QMatrix4x4


class Location(NamedTuple):
    x: float
    y: float
    z: float


class Rotation(NamedTuple):
    azimuth: float
    elevation: float


class Bounds(NamedTuple):
    min_x: float
    min_y: float
    min_z: float
    max_x: float
    max_y: float
    max_z: float


class Camera(QObject):
    """A class to hold the state information of the camera."""

    # Signals
    # The intrinsic or extrinsic matrix changed.
    transform_changed = Signal()
    # Camera internal state (FOV/aspect/projection/clipping) changed.
    intrinsics_changed = Signal()
    # Camera external state (location/rotation) changed.
    extrinsics_changed = Signal()
    # The camera moved
    location_changed = Signal()
    # The camera rotated
    rotation_changed = Signal()

    # Private variables
    _bounds: Bounds
    # Extrinsic attrs
    _location: Location
    _rotation: Rotation
    # Matrix
    _intrinsic_matrix: QMatrix4x4
    _extrinsic_matrix: Optional[QMatrix4x4]

    __slots__ = (
        "_bounds",
        "_location",
        "_rotation",
        "_intrinsic_matrix",
        "_extrinsic_matrix",
    )

    def __init__(self, parent: QObject = None):
        super().__init__(parent)
        self._bounds = Bounds(
            -1_000_000_000,
            -1_000_000_000,
            -1_000_000_000,
            1_000_000_000,
            1_000_000_000,
            1_000_000_000,
        )
        self._location = Location(0.0, 0.0, 0.0)
        self._rotation = Rotation(0.0, 0.0)

        self._intrinsic_matrix = QMatrix4x4()
        self._extrinsic_matrix = None

    def _clamp_location(self, location: Location) -> Location:
        return Location(
            min(max(self._bounds.min_x, location.x), self._bounds.max_x),
            min(max(self._bounds.min_y, location.y), self._bounds.max_y),
            min(max(self._bounds.min_z, location.z), self._bounds.max_z),
        )

    @property
    def location(self) -> Location:
        """The location of the camera. (x, y, z)"""
        return self._location

    @location.setter
    def location(self, location: Location):
        """Set the location of the camera. (x, y, z)."""

        # Clamp location to the bounds.
        location = self._clamp_location(location)

        if location != self._location:
            self._location = location
            self._extrinsic_matrix = None
            self.location_changed.emit()
            self.extrinsics_changed.emit()
            self.transform_changed.emit()

    def _clamp_rotation(self, rotation: Rotation) -> Rotation:
        return Rotation(
            rotation.azimuth
            if -180 <= rotation.azimuth < 180
            else ((rotation.azimuth + 180) % 360) - 180,
            min(max(-90.0, rotation.elevation), 90.0),
        )

    @property
    def rotation(self) -> Rotation:
        """The rotation of the camera. (azimuth/yaw, elevation/pitch).
        This should behave the same as how Minecraft handles it.
        """
        return self._rotation

    @rotation.setter
    def rotation(self, rotation: Rotation):
        """Set the rotation of the camera. (azimuth/yaw, elevation/pitch).
        azimuth (-180 to 180), elevation (-90 to 90)
        This should behave the same as how Minecraft handles it."""

        # Clamp rotation to the bounds
        rotation = self._clamp_rotation(rotation)

        if rotation != self._rotation:
            self._rotation = rotation
            self._extrinsic_matrix = None
            self.rotation_changed.emit()
            self.extrinsics_changed.emit()
            self.transform_changed.emit()

    def set_extrinsics(self, location: Location, rotation: Rotation):
        """Set the camera location and rotation in one property."""
        location = self._clamp_location(location)
        rotation = self._clamp_rotation(rotation)
        location_changed = location != self._location
        rotation_changed = rotation != self._rotation

        if location_changed:
            self._location = location
        if rotation_changed:
            self._rotation = rotation

        if location_changed or rotation_changed:
            self._extrinsic_matrix = None

            if location_changed:
                self.location_changed.emit()

            if rotation_changed:
                self.rotation_changed.emit()

            self.extrinsics_changed.emit()
            self.transform_changed.emit()

    def set_perspective_projection(
        self,
        vertical_fov: float,
        aspect_ratio: float,
        near_plane: float,
        far_plane: float,
    ):
        """Set the projection to perspective with the given settings."""
        self._intrinsic_matrix.setToIdentity()
        self._intrinsic_matrix.perspective(
            vertical_fov, aspect_ratio, near_plane, far_plane
        )
        self.intrinsics_changed.emit()
        self.transform_changed.emit()

    def set_ortho_projection(
        self,
        left: float,
        right: float,
        bottom: float,
        top: float,
        near_plane: float,
        far_plane: float,
    ):
        """Set the projection to orthographic with the given settings."""
        self._intrinsic_matrix.setToIdentity()
        self._intrinsic_matrix.ortho(left, right, bottom, top, near_plane, far_plane)
        self.intrinsics_changed.emit()
        self.transform_changed.emit()

    @property
    def intrinsic_matrix(self) -> QMatrix4x4:
        """The matrix storing all intrinsic parameters (FOV/aspect/projection/clipping)"""
        return self._intrinsic_matrix

    @property
    def extrinsic_matrix(self) -> QMatrix4x4:
        """The matrix storing all extrinsic parameters (location/rotation)"""
        if self._extrinsic_matrix is None:
            self._extrinsic_matrix = QMatrix4x4()
            location = self._location
            rotation = self._rotation
            self._extrinsic_matrix.rotate(rotation.elevation, 1, 0, 0)
            self._extrinsic_matrix.rotate(rotation.azimuth, 0, 1, 0)
            self._extrinsic_matrix.translate(-location.x, -location.y, -location.z)
        return self._extrinsic_matrix



from __future__ import annotations
from typing import Optional, Union

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QVBoxLayout

from amulet_team_main_window.models.view import View


ViewType = Union[QWidget, View]


class ViewContainer(QWidget):
    """A widget to contain a View"""

    # In its basic form this could be implemented with a stacked widget
    # but I would like to look into support for multiple active views
    # and the ability to move views around hence the need for a container

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        f: Qt.WindowType = Qt.WindowType.Widget,
    ):
        super().__init__(parent, f)
        self._uid = None
        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._view: Optional[ViewType] = None

    @property
    def view_cls(self) -> Optional[str]:
        return self._view.__class__

    def get_view(self) -> Optional[ViewType]:
        return self._view

    def swap_view(self, view: ViewType) -> Optional[ViewType]:
        """
        Swap the contained view.
        This orphans and returns the old view.
        It is the job of the caller to either hide and store it for later use or destroy it.

        :param view: The view instance to set. This must be an instance of QWidget or a subclass of QWidget and View.
        :return: The old orphaned view instance.
        """
        if not isinstance(view, View):
            raise TypeError("view must be an instance of View")
        if not isinstance(view, QWidget):
            raise TypeError("view must be an instance of QWidget")
        if view.parent() is not None:
            raise RuntimeError("view must be an orphan.")

        old_view = self._view
        if old_view is not None:
            old_view.disable_view()
            self._layout.removeWidget(old_view)
            # Orphan the widget
            old_view.setParent(None)
        self._view = view
        self._layout.addWidget(view)
        return old_view



from __future__ import annotations

from PySide6.QtCore import Qt

from .settings import SettingsPage

from amulet_team_main_window.api import add_static_toolbar_button


_windows: list = []


def load_plugin():
    add_static_toolbar_button(
        "amulet_team:settings", "settings.svg", "Settings", _on_click
    )


def _on_click(self):
    settings = SettingsPage()
    settings.setWindowModality(Qt.WindowModality.ApplicationModal)
    settings.showNormal()
    self._windows.append(settings)



import os

from amulet_editor.data import system


def save_directories() -> list[str]:
    """
    Returns a list of paths to all default Minecraft save
    directories on the current device.
    """

    directories: list[str] = []

    if system.is_windows():
        directories.extend(
            (
                os.path.join(
                    os.getenv("APPDATA"),
                    ".minecraft",
                    "saves",
                ),
                os.path.join(
                    os.getenv("LOCALAPPDATA"),
                    "Packages",
                    "Microsoft.MinecraftUWP_8wekyb3d8bbwe",
                    "LocalState",
                    "games",
                    "com.mojang",
                    "minecraftWorlds",
                ),
            )
        )
    elif system.is_mac():
        directories.extend(
            (
                os.path.join(
                    os.path.expanduser("~"),
                    "Library",
                    "Application Support",
                    "minecraft",
                    "saves",
                ),
            )
        )
    elif system.is_linux():
        directories.extend(
            (
                os.path.join(
                    os.path.expanduser("~"),
                    ".minecraft",
                    "saves",
                ),
            )
        )

    return list(filter(os.path.isdir, directories))



from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QFileDialog, QApplication
from amulet_editor.data.build import get_resource
from ._open_world import Ui_OpenWorldPage
from amulet_editor.application._cli import spawn_process
from amulet_editor.data.project import get_level


class OpenWorldPage(Ui_OpenWorldPage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.btn_back.setIcon(QIcon(get_resource("icons/tabler/arrow-left.svg")))
        self.load_file_button.clicked.connect(self.open_file)
        self.load_directory_button.clicked.connect(self.open_dir)

    def open_file(self):
        dialog = QFileDialog(self)
        dialog.setFileMode(QFileDialog.FileMode.ExistingFile)
        dialog.setNameFilter("Minecraft files (*.*)")
        dialog.setViewMode(QFileDialog.ViewMode.Detail)
        if dialog.exec():
            file, *_ = dialog.selectedFiles()
            spawn_process(file)
            if get_level() is None:
                QApplication.quit()

    def open_dir(self):
        dialog = QFileDialog(self)
        dialog.setFileMode(QFileDialog.FileMode.Directory)
        if dialog.exec():
            file, *_ = dialog.selectedFiles()
            spawn_process(file)
            if get_level() is None:
                QApplication.quit()



import os
from functools import partial
from typing import Callable, Optional

import amulet
from amulet_editor.data import minecraft, paths
from amulet_editor.models.generic import Observer
from amulet_editor.models.minecraft import LevelData
from amulet_editor.models.widgets import AIconButton
from amulet_editor.tools.startup._models import Menu, ProjectData
from amulet_editor.tools.startup._widgets import QLevelSelectionCard
from amulet_editor.tools.startup.panels._world_selection import WorldSelectionPanel
from PySide6.QtCore import QCoreApplication, QObject, QSize
from PySide6.QtWidgets import (
    QApplication,
    QFileDialog,
    QFrame,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QVBoxLayout,
    QWidget,
)


class ImportLevelMenu(QObject):
    def __init__(self, set_panel: Callable[[Optional[QWidget]], None]):
        super().__init__()

        self.project_directory = paths.project_directory()

        self.level_directory: Optional[str] = None
        self.project_data: Optional[ProjectData] = None
        self.set_panel = set_panel

        self._enable_next = Observer(bool)

        self._widget = ImportLevelWidget()
        self._widget.btn_import_level.clicked.connect(partial(self.import_level))
        self._widget.crd_select_level.clicked.connect(partial(self.select_level))

        self._world_selection_panel = WorldSelectionPanel()
        self._world_selection_panel.level_data.connect(self.set_level)

        QApplication.instance().focusChanged.connect(self.check_focus)

    @property
    def title(self) -> str:
        return "New Project"

    @property
    def enable_next(self) -> Observer:
        return self._enable_next

    def navigated(self, destination):
        pass

    def widget(self) -> QWidget:
        return self._widget

    def next_menu(self) -> Optional[Menu]:
        return None

    def import_level(self):
        path = QFileDialog.getExistingDirectory(
            None,
            "Select Minecraft World",
            os.path.realpath(minecraft.save_directories()[0]),
            QFileDialog.Option.ShowDirsOnly,
        )
        self._widget.btn_import_level.setChecked(False)

        if os.path.exists(os.path.join(path, "level.dat")):
            path = str(os.path.sep).join(path.split("/"))

            level_data = LevelData(amulet.load_format(path))
            self.set_level(level_data)

    def select_level(self):
        if self._widget.crd_select_level.isChecked():
            self.set_panel(self._world_selection_panel)
        else:
            self.set_panel(None)

    def set_level(self, level_data: LevelData):
        self.project_data.level_directory = level_data.path

        self._widget.lne_import_level.setText(level_data.path)
        self._widget.crd_select_level.setLevel(level_data)

        self.enable_next.emit(True)

    def set_project_data(self, project_data: ProjectData):
        self.project_data = project_data

    def check_focus(self, old: Optional[QWidget], new: Optional[QWidget]):
        alternate_focus = [self._widget.lne_import_level]

        if new in alternate_focus:
            self.uncheck_level_card()

    def uncheck_level_card(self):
        self._widget.crd_select_level.setChecked(False)
        self._widget.crd_select_level.clicked.emit()


class ImportLevelWidget(QWidget):
    def __init__(self):
        super().__init__()

        self.setupUi()

    def setupUi(self):
        # Create 'Select Level' frame
        self.lbl_select_level = QLabel(self)
        self.lbl_select_level.setProperty("color", "on_primary")

        lyt_import_level = QHBoxLayout(self)

        self.frm_import_level = QFrame(self)
        self.frm_import_level.setFrameShape(QFrame.Shape.NoFrame)
        self.frm_import_level.setFrameShadow(QFrame.Shadow.Raised)
        self.frm_import_level.setLayout(lyt_import_level)

        self.lne_import_level = QLineEdit(self.frm_import_level)
        self.lne_import_level.setFixedHeight(25)
        self.lne_import_level.setProperty("color", "on_surface")
        self.lne_import_level.setReadOnly(True)

        self.btn_import_level = AIconButton("folder.svg", self.frm_import_level)
        self.btn_import_level.setCheckable(True)
        self.btn_import_level.setFixedSize(QSize(27, 27))
        self.btn_import_level.setIconSize(QSize(15, 15))
        self.btn_import_level.setProperty("backgroundColor", "primary")

        lyt_import_level.addWidget(self.lne_import_level)
        lyt_import_level.addWidget(self.btn_import_level)
        lyt_import_level.setContentsMargins(0, 0, 0, 0)
        lyt_import_level.setSpacing(5)

        self.crd_select_level = QLevelSelectionCard(self)
        self.crd_select_level.setCheckable(True)
        self.crd_select_level.setFixedHeight(105)
        self.crd_select_level.setProperty("backgroundColor", "primary")
        self.crd_select_level.setProperty("border", "surface")
        self.crd_select_level.setProperty("borderRadiusVisible", True)

        # Create 'Page Content' layout
        layout = QVBoxLayout(self)
        layout.addSpacing(10)
        layout.addWidget(self.lbl_select_level)
        layout.addWidget(self.frm_import_level)
        layout.addWidget(self.crd_select_level)
        layout.addStretch()
        layout.setSpacing(5)

        self.setProperty("backgroundColor", "background")
        self.setLayout(layout)

        self.retranslateUi()

    def retranslateUi(self):
        # Disable formatting to condense tranlate functions
        # fmt: off
        self.lbl_select_level.setText(QCoreApplication.translate("NewProjectTypePage", "Select World", None))
        # fmt: on
